import hashlib
import json
import pyckb.bech32
import pyckb.config
import pyckb.ecdsa
import pyckb.molecule
import pyckb.secp256k1
import typing

# https://github.com/nervosnetwork/rfcs/blob/master/rfcs/0022-transaction-structure/0022-transaction-structure.md
# The Type ID code cell uses a special type script hash, which is just the ascii codes in hex of the text TYPE_ID.
type_id_code_hash = bytearray.fromhex('00000000000000000000000000000000000000000000000000545950455f4944')
type_id_hash_type = 1

# Specifies how the script code_hash is used to match the script code and how to run the code.
script_hash_type_data = 0
script_hash_type_type = 1
script_hash_type_data1 = 2
script_hash_type_data2 = 4


def hash(data: bytearray) -> bytearray:
    return bytearray(hashlib.blake2b(data, digest_size=32, person=b'ckb-default-hash').digest())


class PriKey:
    def __init__(self, n: int) -> None:
        self.n = n

    def __repr__(self) -> str:
        return json.dumps(self.json())

    def __eq__(self, other: typing.Self) -> bool:
        return self.n == other.n

    def json(self) -> typing.Dict:
        return {
            'n': f'0x{self.n:064x}',
        }

    def molecule(self) -> bytearray:
        return bytearray(self.n.to_bytes(32))

    @classmethod
    def molecule_decode(cls, data: bytearray) -> typing.Self:
        assert len(data) == 32
        return PriKey(int.from_bytes(data))

    def pubkey(self):
        pubkey = pyckb.secp256k1.G * pyckb.secp256k1.Fr(self.n)
        return PubKey(pubkey.x.x, pubkey.y.x)

    def sign(self, data: bytearray) -> bytearray:
        assert len(data) == 32
        m = pyckb.secp256k1.Fr(int.from_bytes(data))
        r, s, v = pyckb.ecdsa.sign(pyckb.secp256k1.Fr(self.n), m)
        return bytearray(r.x.to_bytes(32)) + bytearray(s.x.to_bytes(32)) + bytearray([v])


class PubKey:
    def __init__(self, x: int, y: int) -> None:
        # The public key must be on the curve.
        _ = pyckb.secp256k1.Pt(pyckb.secp256k1.Fq(x), pyckb.secp256k1.Fq(y))
        self.x = x
        self.y = y

    def __repr__(self) -> str:
        return json.dumps(self.json())

    def __eq__(self, other: typing.Self) -> bool:
        return all([
            self.x == other.x,
            self.y == other.y,
        ])

    def json(self) -> typing.Dict:
        return {
            'x': f'0x{self.x:064x}',
            'y': f'0x{self.y:064x}'
        }

    def pt(self) -> pyckb.secp256k1.Pt:
        return pyckb.secp256k1.Pt(pyckb.secp256k1.Fq(self.x), pyckb.secp256k1.Fq(self.y))

    @classmethod
    def pt_decode(cls, data: pyckb.secp256k1.Pt) -> typing.Self:
        return PubKey(data.x.x, data.y.x)

    def sec(self) -> bytearray:
        # The Standards of Efficient Cryptography (SEC) encoding is used to serialize ECDSA public keys. Public keys in
        # Bitcoin are ECDSA points consisting of two coordinates (x,y). x and y may be smaller than 32 bytes in which
        # case they must be padded with zeros to 32 bytes (H/T Coding Enthusiast).
        r = bytearray()
        if self.y & 1 == 0:
            r.append(0x02)
        else:
            r.append(0x03)
        r.extend(self.x.to_bytes(32))
        return r

    @classmethod
    def sec_decode(cls, data: bytearray) -> typing.Self:
        p = data[0]
        assert p in [0x02, 0x03, 0x04]
        x = int.from_bytes(data[1:33])
        if p == 0x04:
            y = int.from_bytes(data[33:65])
        else:
            y_x_y = x * x * x + pyckb.secp256k1.A.x * x + pyckb.secp256k1.B.x
            y = pow(y_x_y, (pyckb.secp256k1.P + 1) // 4, pyckb.secp256k1.P)
            if y & 1 != p - 2:
                y = -y % pyckb.secp256k1.P
        return PubKey(x, y)


class Script:
    def __init__(self, code_hash: bytearray, hash_type: int, args: bytearray) -> None:
        assert len(code_hash) == 32
        assert hash_type in [
            script_hash_type_data,
            script_hash_type_type,
            script_hash_type_data1,
            script_hash_type_data2,
        ]
        self.code_hash = code_hash
        self.hash_type = hash_type
        self.args = args

    def __repr__(self) -> str:
        return json.dumps(self.json())

    def __eq__(self, other: typing.Self) -> bool:
        return all([
            self.code_hash == other.code_hash,
            self.hash_type == other.hash_type,
            self.args == other.args,
        ])

    def hash(self) -> bytearray:
        return hash(self.molecule())

    def json(self) -> typing.Dict:
        return {
            'code_hash': f'0x{self.code_hash.hex()}',
            'hash_type': {
                script_hash_type_data: 'data',
                script_hash_type_type: 'type',
                script_hash_type_data1: 'data1',
                script_hash_type_data2: 'data2',
            }[self.hash_type],
            'args': f'0x{self.args.hex()}',
        }

    @classmethod
    def json_decode(cls, data: typing.Dict) -> typing.Self:
        return Script(
            bytearray.fromhex(data['code_hash'][2:]),
            {
                'data': script_hash_type_data,
                'type': script_hash_type_type,
                'data1': script_hash_type_data1,
                'data2': script_hash_type_data2,
            }[data['hash_type']],
            bytearray.fromhex(data['args'][2:]),
        )

    def molecule(self) -> bytearray:
        return pyckb.molecule.encode_dynvec([
            pyckb.molecule.Byte32(self.code_hash).molecule(),
            pyckb.molecule.Byte(self.hash_type).molecule(),
            pyckb.molecule.Bytes(self.args).molecule()
        ])

    @classmethod
    def molecule_decode(cls, data: bytearray) -> typing.Self:
        result = pyckb.molecule.decode_dynvec(data)
        return Script(
            pyckb.molecule.Byte32.molecule_decode(result[0]),
            pyckb.molecule.Byte.molecule_decode(result[1]),
            pyckb.molecule.Bytes.molecule_decode(result[2]),
        )


def address_encode(script: Script) -> str:
    # See: https://github.com/nervosnetwork/rfcs/blob/master/rfcs/0021-ckb-address-format/0021-ckb-address-format.md
    # See: https://github.com/rev-chaos/ckb-address-demo/blob/master/ckb_addr_test.py
    payload = bytearray()
    payload.append(0x00)
    # Append secp256k1 code hash
    payload.extend(script.code_hash)
    payload.append(script.hash_type)
    payload.extend(script.args)
    return pyckb.bech32.encode(pyckb.config.current.hrp, payload)


def address_decode(addr: str) -> Script:
    payload = pyckb.bech32.decode(pyckb.config.current.hrp, addr)
    assert payload[0] == 0
    code_hash = payload[1:33]
    hash_type = payload[33]
    args = payload[34:]
    return Script(code_hash, hash_type, args)


class OutPoint:
    def __init__(self, tx_hash: bytearray, index: int) -> None:
        assert len(tx_hash) == 32
        self.tx_hash = tx_hash
        self.index = index

    def __repr__(self) -> str:
        return json.dumps(self.json())

    def __eq__(self, other: typing.Self) -> bool:
        return all([
            self.tx_hash == other.tx_hash,
            self.index == other.index,
        ])

    def json(self) -> typing.Dict:
        return {
            'tx_hash': '0x' + self.tx_hash.hex(),
            'index': hex(self.index),
        }

    @classmethod
    def json_decode(cls, data: typing.Dict) -> typing.Self:
        return OutPoint(
            bytearray.fromhex(data['tx_hash'][2:]),
            int(data['index'], 16),
        )

    def molecule(self) -> bytearray:
        return pyckb.molecule.encode_seq([
            pyckb.molecule.Byte32(self.tx_hash).molecule(),
            pyckb.molecule.U32(self.index).molecule(),
        ])

    @classmethod
    def molecule_decode(cls, data: bytearray) -> typing.Self:
        result = pyckb.molecule.decode_seq(data, [
            pyckb.molecule.Byte32.molecule_size(),
            pyckb.molecule.U32.molecule_size(),
        ])
        return OutPoint(
            pyckb.molecule.Byte32.molecule_decode(result[0]),
            pyckb.molecule.U32.molecule_decode(result[1])
        )

    @classmethod
    def molecule_size(cls) -> int:
        return pyckb.molecule.Byte32.molecule_size() + pyckb.molecule.U32.molecule_size()


class CellInput:
    def __init__(self, since: int, previous_output: OutPoint) -> None:
        self.since = since
        self.previous_output = previous_output

    def __repr__(self) -> str:
        return json.dumps(self.json())

    def __eq__(self, other: typing.Self) -> bool:
        return all([
            self.since == other.since,
            self.previous_output == other.previous_output,
        ])

    def json(self) -> typing.Dict:
        return {
            'since': hex(self.since),
            'previous_output': self.previous_output.json()
        }

    @classmethod
    def json_decode(cls, data: typing.Dict) -> typing.Self:
        return CellInput(
            int(data['since'], 16),
            OutPoint.json_decode(data['previous_output']),
        )

    def molecule(self) -> bytearray:
        return pyckb.molecule.encode_seq([
            pyckb.molecule.U64(self.since).molecule(),
            self.previous_output.molecule(),
        ])

    @classmethod
    def molecule_decode(cls, data: bytearray) -> typing.Self:
        result = pyckb.molecule.decode_seq(data, [
            pyckb.molecule.U64.molecule_size(),
            OutPoint.molecule_size(),
        ])
        return CellInput(
            pyckb.molecule.U64.molecule_decode(result[0]),
            OutPoint.molecule_decode(result[1]),
        )

    @classmethod
    def molecule_size(cls) -> int:
        return pyckb.molecule.U64.molecule_size() + OutPoint.molecule_size()


class CellOutput:
    def __init__(self, capacity: int, lock: Script, type: Script | None) -> None:
        self.capacity = capacity
        self.lock = lock
        self.type = type

    def __repr__(self) -> str:
        return json.dumps(self.json())

    def __eq__(self, other: typing.Self) -> bool:
        return all([
            self.capacity == other.capacity,
            self.lock == other.lock,
            self.type == other.type,
        ])

    def json(self) -> typing.Dict:
        return {
            'capacity': hex(self.capacity),
            'lock': self.lock.json(),
            'type': self.type.json() if self.type else None
        }

    @classmethod
    def json_decode(cls, data: typing.Dict) -> typing.Self:
        return CellOutput(
            int(data['capacity'], 16),
            Script.json_decode(data['lock']),
            Script.json_decode(data['type']) if data['type'] else None,
        )

    def molecule(self) -> bytearray:
        return pyckb.molecule.encode_dynvec([
            pyckb.molecule.U64(self.capacity).molecule(),
            self.lock.molecule(),
            self.type.molecule() if self.type else bytearray(),
        ])

    @classmethod
    def molecule_decode(cls, data: bytearray) -> typing.Self:
        result = pyckb.molecule.decode_dynvec(data)
        return CellOutput(
            pyckb.molecule.U64.molecule_decode(result[0]),
            Script.molecule_decode(result[1]),
            Script.molecule_decode(result[2]) if result[2] else None
        )


class CellDep:
    def __init__(self, out_point: OutPoint, dep_type: int) -> None:
        self.out_point = out_point
        self.dep_type = dep_type

    def __repr__(self) -> str:
        return json.dumps(self.json())

    def __eq__(self, other: typing.Self) -> bool:
        return all([
            self.out_point == other.out_point,
            self.dep_type == other.dep_type,
        ])

    @classmethod
    def conf_decode(cls, data: typing.Dict) -> typing.Self:
        return CellDep(OutPoint(data.out_point.tx_hash, data.out_point.index), data.dep_type)

    def json(self) -> typing.Dict:
        return {
            'out_point': self.out_point.json(),
            'dep_type': {0: 'code', 1: 'dep_group'}[self.dep_type]
        }

    @classmethod
    def json_decode(cls, data: typing.Dict) -> typing.Self:
        return CellDep(
            OutPoint.json_decode(data['out_point']),
            {'code': 0, 'dep_group': 1}[data['dep_type']],
        )

    def molecule(self) -> bytearray:
        return pyckb.molecule.encode_seq([
            self.out_point.molecule(),
            pyckb.molecule.Byte(self.dep_type).molecule(),
        ])

    @classmethod
    def molecule_decode(cls, data: bytearray) -> typing.Self:
        result = pyckb.molecule.decode_seq(data, [
            OutPoint.molecule_size(),
            pyckb.molecule.Byte.molecule_size(),
        ])
        return CellDep(
            OutPoint.molecule_decode(result[0]),
            pyckb.molecule.Byte.molecule_decode(result[1]),
        )

    @classmethod
    def molecule_size(cls) -> int:
        return OutPoint.molecule_size() + pyckb.molecule.Byte.molecule_size()


class RawTransaction:
    def __init__(
        self,
        version: int,
        cell_deps: list[CellDep],
        header_deps: list[bytearray],
        inputs: list[CellInput],
        outputs: list[CellOutput],
        outputs_data: list[bytearray]
    ) -> None:
        self.version = version
        self.cell_deps = cell_deps
        self.header_deps = header_deps
        self.inputs = inputs
        self.outputs = outputs
        self.outputs_data = outputs_data

    def __repr__(self) -> str:
        return json.dumps(self.json())

    def __eq__(self, other: typing.Self) -> bool:
        return all([
            self.version == other.version,
            self.cell_deps == other.cell_deps,
            self.header_deps == other.header_deps,
            self.inputs == other.inputs,
            self.outputs == other.outputs,
            self.outputs_data == other.outputs_data,
        ])

    def hash(self) -> bytearray:
        return hash(self.molecule())

    def json(self) -> typing.Dict:
        return {
            'version': hex(self.version),
            'cell_deps': [e.json() for e in self.cell_deps],
            'header_deps': ['0x' + e.hex() for e in self.header_deps],
            'inputs': [e.json() for e in self.inputs],
            'outputs': [e.json() for e in self.outputs],
            'outputs_data': ['0x' + e.hex() for e in self.outputs_data],
        }

    @classmethod
    def json_decode(cls, data: typing.Dict) -> typing.Self:
        return RawTransaction(
            int(data['version'], 16),
            [CellDep.json_decode(e) for e in data['cell_deps']],
            [bytearray.fromhex(e[2:]) for e in data['header_deps']],
            [CellInput.json_decode(e) for e in data['inputs']],
            [CellOutput.json_decode(e) for e in data['outputs']],
            [bytearray.fromhex(e[2:]) for e in data['outputs_data']],
        )

    def molecule(self) -> bytearray:
        return pyckb.molecule.encode_dynvec([
            pyckb.molecule.U32(self.version).molecule(),
            pyckb.molecule.encode_fixvec([e.molecule() for e in self.cell_deps]),
            pyckb.molecule.encode_fixvec([pyckb.molecule.Byte32(e).molecule() for e in self.header_deps]),
            pyckb.molecule.encode_fixvec([e.molecule() for e in self.inputs]),
            pyckb.molecule.encode_dynvec([e.molecule() for e in self.outputs]),
            pyckb.molecule.encode_dynvec([pyckb.molecule.Bytes(e).molecule() for e in self.outputs_data])
        ])

    @classmethod
    def molecule_decode(cls, data: bytearray) -> typing.Self:
        result = pyckb.molecule.decode_dynvec(data)
        return RawTransaction(
            pyckb.molecule.U32.molecule_decode(result[0]),
            [CellDep.molecule_decode(e) for e in pyckb.molecule.decode_fixvec(result[1])],
            [pyckb.molecule.Byte32.molecule_decode(e) for e in pyckb.molecule.decode_fixvec(result[2])],
            [CellInput.molecule_decode(e) for e in pyckb.molecule.decode_fixvec(result[3])],
            [CellOutput.molecule_decode(e) for e in pyckb.molecule.decode_dynvec(result[4])],
            [pyckb.molecule.Bytes.molecule_decode(e) for e in pyckb.molecule.decode_dynvec(result[5])]
        )


class Transaction:
    def __init__(self, raw: RawTransaction, witnesses: list[bytearray]) -> None:
        self.raw = raw
        self.witnesses = witnesses

    def __repr__(self) -> str:
        return json.dumps(self.json())

    def __eq__(self, other: typing.Self) -> bool:
        return all([
            self.raw == other.raw,
            self.witnesses == other.witnesses,
        ])

    def json(self) -> typing.Dict:
        r = self.raw.json()
        r['witnesses'] = [f'0x{e.hex()}' for e in self.witnesses]
        return r

    @classmethod
    def json_decode(cls, data: typing.Dict) -> typing.Self:
        return Transaction(
            RawTransaction.json_decode(data),
            [bytearray.fromhex(e[2:]) for e in data['witnesses']],
        )

    def molecule(self) -> bytearray:
        return pyckb.molecule.encode_dynvec([
            self.raw.molecule(),
            pyckb.molecule.encode_dynvec([pyckb.molecule.Bytes(e).molecule() for e in self.witnesses])
        ])

    @classmethod
    def molecule_decode(cls, data: bytearray) -> typing.Self:
        result = pyckb.molecule.decode_dynvec(data)
        return Transaction(
            RawTransaction.molecule_decode(result[0]),
            [pyckb.molecule.Bytes.molecule_decode(e) for e in pyckb.molecule.decode_dynvec(result[1])],
        )


def epoch_encode(e: int, i: int, l: int) -> int:
    assert 0 <= e and e <= 0xffffff
    assert 0 <= i and i <= 0xffff
    assert 0 <= l and l <= 0xffff
    return l << 0x28 | i << 0x18 | e


def epoch_decode(v: int) -> typing.Tuple[int, int, int]:
    e = v & 0xffffff
    i = v >> 0x18 & 0xffff
    l = v >> 0x28 & 0xffff
    return e, i, l


class WitnessArgs:
    def __init__(self, lock: bytearray | None, input_type: bytearray | None, output_type: bytearray | None) -> None:
        self.lock = lock
        self.input_type = input_type
        self.output_type = output_type

    def __repr__(self) -> str:
        return json.dumps(self.json())

    def __eq__(self, other: typing.Self) -> bool:
        return all([
            self.lock == other.lock,
            self.input_type == other.input_type,
            self.output_type == other.output_type,
        ])

    def json(self) -> typing.Dict:
        return {
            'lock': f'0x{self.lock.hex()}' if self.lock else None,
            'input_type': f'0x{self.input_type.hex()}' if self.input_type else None,
            'output_type': f'0x{self.output_type.hex()}' if self.output_type else None,
        }

    def molecule(self) -> bytearray:
        return pyckb.molecule.encode_dynvec([
            pyckb.molecule.Bytes(self.lock).molecule() if self.lock else bytearray(),
            pyckb.molecule.Bytes(self.input_type).molecule() if self.input_type else bytearray(),
            pyckb.molecule.Bytes(self.output_type).molecule() if self.output_type else bytearray(),
        ])

    @classmethod
    def molecule_decode(cls, data: bytearray) -> typing.Self:
        result = pyckb.molecule.decode_dynvec(data)
        return WitnessArgs(
            pyckb.molecule.Bytes.molecule_decode(result[0]) if result[0] else None,
            pyckb.molecule.Bytes.molecule_decode(result[1]) if result[1] else None,
            pyckb.molecule.Bytes.molecule_decode(result[2]) if result[2] else None,
        )


class RawHeader:
    def __init__(
        self,
        version: int,
        compact_target: int,
        timestamp: int,
        number: int,
        epoch: int,
        parent_hash: bytearray,
        transactions_root: bytearray,
        proposals_hash: bytearray,
        extra_hash: bytearray,
        dao: bytearray,
    ) -> None:
        self.version = version
        self.compact_target = compact_target
        self.timestamp = timestamp
        self.number = number
        self.epoch = epoch
        self.parent_hash = parent_hash
        self.transactions_root = transactions_root
        self.proposals_hash = proposals_hash
        self.extra_hash = extra_hash
        self.dao = dao

    def __repr__(self) -> str:
        return json.dumps(self.json())

    def __eq__(self, other: typing.Self) -> bool:
        return all([
            self.version == other.version,
            self.compact_target == other.compact_target,
            self.timestamp == other.timestamp,
            self.number == other.number,
            self.epoch == other.epoch,
            self.parent_hash == other.parent_hash,
            self.transactions_root == other.transactions_root,
            self.proposals_hash == other.proposals_hash,
            self.extra_hash == other.extra_hash,
            self.dao == other.dao,
        ])

    def json(self) -> typing.Dict:
        return {
            'version': f'0x{hex(self.version)}',
            'compact_target': f'0x{hex(self.compact_target)}',
            'timestamp': f'0x{hex(self.timestamp)}',
            'number': f'0x{hex(self.number)}',
            'epoch': f'0x{hex(self.epoch)}',
            'parent_hash': f'0x{self.parent_hash.hex()}',
            'transactions_root': f'0x{self.transactions_root.hex()}',
            'proposals_hash': f'0x{self.proposals_hash.hex()}',
            'extra_hash': f'0x{self.extra_hash.hex()}',
            'dao': f'0x{self.dao.hex()}',
        }

    @classmethod
    def json_decode(cls, data: typing.Dict) -> typing.Self:
        return RawHeader(
            version=int(data['version'], 16),
            compact_target=int(data['compact_target'], 16),
            timestamp=int(data['timestamp'], 16),
            number=int(data['number'], 16),
            epoch=int(data['epoch'], 16),
            parent_hash=bytearray.fromhex(data['parent_hash'][2:]),
            transactions_root=bytearray.fromhex(data['transactions_root'][2:]),
            proposals_hash=bytearray.fromhex(data['proposals_hash'][2:]),
            extra_hash=bytearray.fromhex(data['extra_hash'][2:]),
            dao=bytearray.fromhex(data['dao'][2:]),
        )

    def molecule(self) -> bytearray:
        return pyckb.molecule.encode_seq([
            pyckb.molecule.U32(self.version).molecule(),
            pyckb.molecule.U32(self.compact_target).molecule(),
            pyckb.molecule.U64(self.timestamp).molecule(),
            pyckb.molecule.U64(self.number).molecule(),
            pyckb.molecule.U64(self.epoch).molecule(),
            pyckb.molecule.Byte32(self.parent_hash).molecule(),
            pyckb.molecule.Byte32(self.transactions_root).molecule(),
            pyckb.molecule.Byte32(self.proposals_hash).molecule(),
            pyckb.molecule.Byte32(self.extra_hash).molecule(),
            pyckb.molecule.Byte32(self.dao).molecule(),
        ])

    @classmethod
    def molecule_decode(cls, data: bytearray) -> typing.Self:
        result = pyckb.molecule.decode_seq(data, [
            pyckb.molecule.U32.molecule_size(),
            pyckb.molecule.U32.molecule_size(),
            pyckb.molecule.U64.molecule_size(),
            pyckb.molecule.U64.molecule_size(),
            pyckb.molecule.U64.molecule_size(),
            pyckb.molecule.Byte32.molecule_size(),
            pyckb.molecule.Byte32.molecule_size(),
            pyckb.molecule.Byte32.molecule_size(),
            pyckb.molecule.Byte32.molecule_size(),
            pyckb.molecule.Byte32.molecule_size(),
        ])
        return RawHeader(
            pyckb.molecule.U32.molecule_decode(result[0]),
            pyckb.molecule.U32.molecule_decode(result[1]),
            pyckb.molecule.U64.molecule_decode(result[2]),
            pyckb.molecule.U64.molecule_decode(result[3]),
            pyckb.molecule.U64.molecule_decode(result[4]),
            pyckb.molecule.Byte32.molecule_decode(result[5]),
            pyckb.molecule.Byte32.molecule_decode(result[6]),
            pyckb.molecule.Byte32.molecule_decode(result[7]),
            pyckb.molecule.Byte32.molecule_decode(result[8]),
            pyckb.molecule.Byte32.molecule_decode(result[9]),
        )

    @classmethod
    def molecule_size(cls) -> int:
        return sum([
            pyckb.molecule.U32.molecule_size(),
            pyckb.molecule.U32.molecule_size(),
            pyckb.molecule.U64.molecule_size(),
            pyckb.molecule.U64.molecule_size(),
            pyckb.molecule.U64.molecule_size(),
            pyckb.molecule.Byte32.molecule_size(),
            pyckb.molecule.Byte32.molecule_size(),
            pyckb.molecule.Byte32.molecule_size(),
            pyckb.molecule.Byte32.molecule_size(),
            pyckb.molecule.Byte32.molecule_size(),
        ])


class Header:
    def __init__(self, raw: RawHeader, nonce: int) -> None:
        self.raw = raw
        self.nonce = nonce

    def __repr__(self) -> str:
        return json.dumps(self.json())

    def __eq__(self, other: typing.Self) -> bool:
        return all([
            self.raw == other.raw,
            self.nonce == other.nonce,
        ])

    def hash(self) -> bytearray:
        return hash(self.molecule())

    def json(self) -> typing.Dict:
        r = self.raw.json()
        r['nonce'] = f'0x{hex(self.nonce)}'
        return r

    @classmethod
    def json_decode(cls, data: typing.Dict) -> typing.Self:
        return Header(
            RawHeader.json_decode(data),
            int(data['nonce'], 16),
        )

    def molecule(self) -> bytearray:
        return pyckb.molecule.encode_seq([
            self.raw.molecule(),
            pyckb.molecule.U128(self.nonce).molecule(),
        ])

    @classmethod
    def molecule_decode(cls, data: bytearray) -> typing.Self:
        result = pyckb.molecule.decode_seq(data, [
            RawHeader.molecule_size(),
            pyckb.molecule.U128.molecule_size(),
        ])
        return Header(
            RawHeader.molecule_decode(result[0]),
            pyckb.molecule.U128.molecule_decode(result[1]),
        )

    @classmethod
    def molecule_size(cls) -> int:
        return RawHeader.molecule_size() + pyckb.molecule.U128.molecule_size()


def dao_encode(c: int, ar: int, s: int, u: int) -> bytearray:
    # CKB's block header has a particular field named dao containing auxiliary information for Nervos DAO's use.
    # https://github.com/nervosnetwork/rfcs/blob/master/rfcs/0023-dao-deposit-withdraw/0023-dao-deposit-withdraw.md
    r = bytearray()
    r.extend(bytearray(c.to_bytes(8, 'little')))
    r.extend(bytearray(ar.to_bytes(8, 'little')))
    r.extend(bytearray(s.to_bytes(8, 'little')))
    r.extend(bytearray(u.to_bytes(8, 'little')))
    return r


def dao_decode(d: bytearray) -> typing.Tuple[int, int, int, int]:
    c = int.from_bytes(d[0x00:0x08], 'little')
    ar = int.from_bytes(d[0x08:0x10], 'little')
    s = int.from_bytes(d[0x10:0x18], 'little')
    u = int.from_bytes(d[0x18:0x20], 'little')
    return c, ar, s, u


class UncleBlock:
    def __init__(self, header: Header, proposals: typing.List[bytearray]) -> None:
        self.header = header
        self.proposals = proposals

    def __repr__(self) -> str:
        return json.dumps(self.json())

    def __eq__(self, other: typing.Self) -> bool:
        return all([
            self.header == other.header,
            self.proposals == other.proposals,
        ])

    def json(self) -> typing.Dict:
        return {
            'header': self.header.json(),
            'proposals': [f'0x{e.hex()}' for e in self.proposals],
        }

    @classmethod
    def json_decode(cls, data: typing.Dict) -> typing.Self:
        return UncleBlock(
            Header.json_decode(data['header']),
            [bytearray.fromhex(e[2:]) for e in data['proposals']]
        )

    def molecule(self) -> bytearray:
        return pyckb.molecule.encode_dynvec([
            self.header.molecule(),
            pyckb.molecule.encode_fixvec(self.proposals),
        ])

    @classmethod
    def molecule_decode(cls, data: bytearray) -> typing.Self:
        result = pyckb.molecule.decode_dynvec(data)
        return UncleBlock(
            Header.molecule_decode(result[0]),
            pyckb.molecule.decode_fixvec(result[1]),
        )


class Block:
    def __init__(
        self,
        header: Header,
        uncles: typing.List[UncleBlock],
        transactions: typing.List[Transaction],
        proposals: typing.List[bytearray],
    ) -> None:
        self.header = header
        self.uncles = uncles
        self.transactions = transactions
        self.proposals = proposals

    def __repr__(self) -> str:
        return json.dumps(self.json())

    def __eq__(self, other: typing.Self) -> bool:
        return all([
            self.header == other.header,
            self.uncles == other.uncles,
            self.transactions == other.transactions,
            self.proposals == other.proposals,
        ])

    def json(self) -> typing.Dict:
        return {
            'header': self.header.json(),
            'uncles': [e.json() for e in self.uncles],
            'transactions': [e.json() for e in self.transactions],
            'proposals': [f'0x{e.hex()}' for e in self.proposals],
        }

    @classmethod
    def json_decode(cls, data: typing.Dict) -> typing.Self:
        return Block(
            Header.json_decode(data['header']),
            [UncleBlock.json_decode(e) for e in data['uncles']],
            [Transaction.json_decode(e) for e in data['transactions']],
            [bytearray.fromhex(e[2:]) for e in data['proposals']],
        )

    def molecule(self) -> bytearray:
        return pyckb.molecule.encode_dynvec([
            self.header.molecule(),
            pyckb.molecule.encode_dynvec([e.molecule() for e in self.uncles]),
            pyckb.molecule.encode_dynvec([e.molecule() for e in self.transactions]),
            pyckb.molecule.encode_fixvec(self.proposals),
        ])

    @classmethod
    def molecule_decode(cls, data: bytearray) -> typing.Self:
        result = pyckb.molecule.decode_dynvec(data)
        return Block(
            Header.molecule_decode(result[0]),
            [UncleBlock.molecule_decode(e) for e in pyckb.molecule.decode_dynvec(result[1])],
            [Transaction.molecule_decode(e) for e in pyckb.molecule.decode_dynvec(result[2])],
            pyckb.molecule.decode_fixvec(result[3]),
        )


class BlockV1:
    def __init__(
        self,
        header: Header,
        uncles: typing.List[UncleBlock],
        transactions: typing.List[Transaction],
        proposals: typing.List[bytearray],
        extension: bytearray,
    ) -> None:
        self.header = header
        self.uncles = uncles
        self.transactions = transactions
        self.proposals = proposals
        self.extension = extension

    def __repr__(self) -> str:
        return json.dumps(self.json())

    def __eq__(self, other: typing.Self) -> bool:
        return all([
            self.header == other.header,
            self.uncles == other.uncles,
            self.transactions == other.transactions,
            self.proposals == other.proposals,
            self.extension == other.extension,
        ])

    def json(self) -> typing.Dict:
        return {
            'header': self.header.json(),
            'uncles': [e.json() for e in self.uncles],
            'transactions': [e.json() for e in self.transactions],
            'proposals': [f'0x{e.hex()}' for e in self.proposals],
            'extension': f'0x{self.extension.hex()}'
        }

    @classmethod
    def json_decode(cls, data: typing.Dict) -> typing.Self:
        return BlockV1(
            Header.json_decode(data['header']),
            [UncleBlock.json_decode(e) for e in data['uncles']],
            [Transaction.json_decode(e) for e in data['transactions']],
            [bytearray.fromhex(e[2:]) for e in data['proposals']],
            bytearray.fromhex(data['extension'][2:]),
        )

    def molecule(self) -> bytearray:
        return pyckb.molecule.encode_dynvec([
            self.header.molecule(),
            pyckb.molecule.encode_dynvec([e.molecule() for e in self.uncles]),
            pyckb.molecule.encode_dynvec([e.molecule() for e in self.transactions]),
            pyckb.molecule.encode_fixvec(self.proposals),
            pyckb.molecule.Bytes(self.extension).molecule(),
        ])

    @classmethod
    def molecule_decode(cls, data: bytearray) -> typing.Self:
        result = pyckb.molecule.decode_dynvec(data)
        return BlockV1(
            Header.molecule_decode(result[0]),
            [UncleBlock.molecule_decode(e) for e in pyckb.molecule.decode_dynvec(result[1])],
            [Transaction.molecule_decode(e) for e in pyckb.molecule.decode_dynvec(result[2])],
            pyckb.molecule.decode_fixvec(result[3]),
            pyckb.molecule.Bytes.molecule_decode(result[4]),
        )


class CellbaseWitness:
    def __init__(self, lock: Script, message: bytearray) -> None:
        self.lock = lock
        self.message = message

    def __repr__(self) -> str:
        return json.dumps(self.json())

    def __eq__(self, other: typing.Self) -> bool:
        return all([
            self.lock == other.lock,
            self.message == other.message,
        ])

    def json(self) -> typing.Dict:
        return {
            'lock': self.lock.json(),
            'message': f'0x{self.message.hex()}',
        }

    @classmethod
    def json_deocde(cls, data: typing.Dict) -> typing.Self:
        return CellbaseWitness(
            Script.json_decode(data['lock']),
            bytearray.fromhex(data['message'][2:]),
        )

    def molecule(self) -> bytearray:
        return pyckb.molecule.encode_dynvec([
            self.lock.molecule(),
            pyckb.molecule.Bytes(self.message).molecule(),
        ])

    @classmethod
    def molecule_decode(cls, data: bytearray) -> typing.Self:
        result = pyckb.molecule.decode_dynvec(data)
        return CellbaseWitness(
            Script.molecule_decode(result[0]),
            pyckb.molecule.Bytes.molecule_decode(result[1]),
        )



import pyckb
import random


def test_addr():
    pyckb.config.upgrade('http://127.0.0.1:8114')
    pyckb.config.current = pyckb.config.develop
    prikey = pyckb.core.PriKey(1)
    pubkey = prikey.pubkey()
    args = pyckb.core.hash(pubkey.sec())[:20]
    script = pyckb.core.Script(
        pyckb.config.current.script.secp256k1_blake160.code_hash,
        pyckb.config.current.script.secp256k1_blake160.hash_type,
        args
    )
    addr = pyckb.core.address_encode(script)
    assert addr == 'ckt1qzda0cr08m85hc8jlnfp3zer7xulejywt49kt2rr0vthywaa50xwsqt4z78ng4yutl5u6xsv27ht6q08mhujf8s2r0n40'
    assert pyckb.core.address_decode(addr) == script
    assert script.hash().hex() == '0b1bae4beaf456349c63c3ce67491fc75a1276d7f9eedd7ea84d6a77f9f3f5f7'
    assert pyckb.core.Script.molecule_decode(script.molecule()) == script


def test_block():
    pyckb.config.current = pyckb.config.mainnet
    block_num = random.randint(1, 11333728)
    block_hex = pyckb.rpc.call('get_block_by_number', [hex(block_num), '0x0'])
    block_bin = bytearray.fromhex(block_hex[2:])
    block = pyckb.core.Block.molecule_decode(block_bin)
    assert block.molecule() == block_bin


def test_block_v1():
    pyckb.config.current = pyckb.config.mainnet
    block_num = random.randint(11333729, int(pyckb.rpc.get_tip_block_number(), 16))
    block_hex = pyckb.rpc.call('get_block_by_number', [hex(block_num), '0x0'])
    block_bin = bytearray.fromhex(block_hex[2:])
    block = pyckb.core.BlockV1.molecule_decode(block_bin)
    assert block.molecule() == block_bin


def test_epoch():
    assert pyckb.core.epoch_decode(0x3690138000093) == (147, 312, 873)
    assert pyckb.core.epoch_encode(147, 312, 873) == 0x3690138000093


def test_pubkey():
    # Double checked by https://pyckb.tools/generator
    prikey = pyckb.core.PriKey(1)
    pubkey = prikey.pubkey()
    assert pubkey.x == 0x79be667ef9dcbbac55a06295ce870b07029bfcdb2dce28d959f2815b16f81798
    assert pubkey.y == 0x483ada7726a3c4655da4fbfc0e1108a8fd17b448a68554199c47d08ffb10d4b8
    assert pubkey.sec().hex() == '0279be667ef9dcbbac55a06295ce870b07029bfcdb2dce28d959f2815b16f81798'
    assert pyckb.core.PubKey.sec_decode(pubkey.sec()) == pubkey


def test_pubkey_hash():
    prikey = pyckb.core.PriKey(1)
    pubkey = prikey.pubkey()
    assert pyckb.core.hash(pubkey.sec())[:20].hex() == '75178f34549c5fe9cd1a0c57aebd01e7ddf9249e'


def test_script():
    script = pyckb.core.Script(
        pyckb.config.current.script.secp256k1_blake160.code_hash,
        pyckb.config.current.script.secp256k1_blake160.hash_type,
        bytearray([0x00, 0x01, 0x02, 0x03])
    )
    assert pyckb.core.Script.molecule_decode(script.molecule()) == script


def test_sign():
    prikey = pyckb.core.PriKey(1)
    prikey.sign(bytearray.fromhex('9bd7e06f3ecf4be0f2fcd2188b23f1b9fcc88e5d4b65a8637b17723bbda3cce8'))


def test_witness_args():
    witness_args = pyckb.core.WitnessArgs(
        bytearray([0x00, 0x01, 0x02, 0x03]),
        bytearray([0x00, 0x01, 0x02, 0x03]),
        None,
    )
    assert pyckb.core.WitnessArgs.molecule_decode(witness_args.molecule()) == witness_args



import argparse
import pyckb

# Transfer ckb to another account. If value is 0, then all assets will be transferred.

parser = argparse.ArgumentParser()
parser.add_argument('--net', type=str, choices=['develop', 'mainnet', 'testnet'], default='develop')
parser.add_argument('--prikey', type=str, help='private key')
parser.add_argument('--to', type=str, help='ckb address')
parser.add_argument('--value', type=float, help='ckb value')
args = parser.parse_args()

if args.net == 'develop':
    pyckb.config.upgrade('http://127.0.0.1:8114')
    pyckb.config.current = pyckb.config.develop
if args.net == 'mainnet':
    pyckb.config.current = pyckb.config.mainnet
if args.net == 'testnet':
    pyckb.config.current = pyckb.config.testnet

user = pyckb.wallet.Wallet(int(args.prikey, 0))
hole = pyckb.core.address_decode(args.to)
if args.value == 0:
    hash = user.transfer_all(hole)
    print(f'0x{hash.hex()}')
    pyckb.rpc.wait(f'0x{hash.hex()}')
if args.value >= 1:
    hash = user.transfer(hole, int(args.value * pyckb.denomination.ckbytes))
    print(f'0x{hash.hex()}')
    pyckb.rpc.wait(f'0x{hash.hex()}')



import argparse
import pyckb

# Calculate address from private key in secp256k1 lock.

parser = argparse.ArgumentParser()
parser.add_argument('--net', type=str, choices=['develop', 'mainnet', 'testnet'], default='develop')
parser.add_argument('--prikey', type=str, help='private key')
args = parser.parse_args()

if args.net == 'develop':
    pyckb.config.upgrade('http://127.0.0.1:8114')
    pyckb.config.current = pyckb.config.develop
if args.net == 'mainnet':
    pyckb.config.current = pyckb.config.mainnet
if args.net == 'testnet':
    pyckb.config.current = pyckb.config.testnet

prikey = pyckb.core.PriKey(int(args.prikey, 0))
pubkey = prikey.pubkey()
args = pyckb.core.hash(pubkey.sec())[:20]
script = pyckb.core.Script(
    pyckb.config.current.script.secp256k1_blake160.code_hash,
    pyckb.config.current.script.secp256k1_blake160.hash_type,
    args
)
addr = pyckb.core.address_encode(script)
print(addr)



# Doc: https://github.com/nervosnetwork/rfcs/blob/master/rfcs/0008-serialization/0008-serialization.md
import typing


def decode_dynvec(data: bytearray) -> typing.List[bytearray]:
    assert len(data) >= 4
    assert len(data) == int.from_bytes(data[0:4], 'little')
    if len(data) == 4:
        return []
    nums = int.from_bytes(data[4:8], 'little') // 4 - 1
    head = []
    for i in range(nums):
        head.append(int.from_bytes(data[i * 4 + 4: i * 4 + 8], 'little'))
    head.append(len(data))
    body = []
    for i in range(nums):
        body.append(data[head[i]:head[i+1]])
    return body


def decode_fixvec(data: bytearray) -> typing.List[bytearray]:
    assert len(data) >= 4
    icnt = int.from_bytes(data[0:4], 'little')
    body = []
    if icnt > 0:
        size = len(data[4:]) // icnt
        for i in range(icnt):
            body.append(data[4+i*size:4+i*size+size])
    return body


def decode_seq(data: bytearray, size: list[int]) -> typing.List[bytearray]:
    r = []
    s = 0
    for n in size:
        r.append(data[s:s+n])
        s += n
    return r


def encode_dynvec(data: list[bytearray]) -> bytearray:
    head = bytearray()
    body = bytearray()
    head_size = 4 + 4 * len(data)
    body_size = 0
    for item in data:
        size = head_size + body_size
        head.extend(bytearray(size.to_bytes(4, 'little')))
        body.extend(item)
        body_size += len(item)
    size = head_size + body_size
    return bytearray(size.to_bytes(4, 'little')) + head + body


def encode_fixvec(data: list[bytearray]) -> bytearray:
    r = bytearray()
    r.extend(len(data).to_bytes(4, 'little'))
    for e in data:
        r.extend(e)
    return r


def encode_seq(data: list[bytearray]) -> bytearray:
    r = bytearray()
    for e in data:
        r.extend(e)
    return r


class Byte:
    def __init__(self, data: int) -> None:
        assert 0 <= data and data <= 0xff
        self.data = data

    def __eq__(self, other) -> bool:
        return self.data == other.data

    def molecule(self) -> bytearray:
        return bytearray([self.data])

    @classmethod
    def molecule_decode(cls, data: bytearray) -> int:
        assert len(data) == 1
        return data[0]

    @classmethod
    def molecule_size(cls) -> int:
        return 1


class Byte32:
    def __init__(self, data: bytearray) -> None:
        assert len(data) == 32
        self.data = data

    def __eq__(self, other) -> bool:
        return self.data == other.data

    def molecule(self) -> bytearray:
        return self.data

    @classmethod
    def molecule_decode(cls, data: bytearray) -> bytearray:
        return data

    @classmethod
    def molecule_size(cls) -> int:
        return 32


class Bytes:
    def __init__(self, data: bytearray) -> None:
        self.data = data

    def __eq__(self, other) -> bool:
        return self.data == other.data

    def molecule(self) -> bytearray:
        r = bytearray()
        r.extend(len(self.data).to_bytes(4, 'little'))
        r.extend(self.data)
        return r

    @classmethod
    def molecule_decode(cls, data: bytearray) -> bytearray:
        l = int.from_bytes(data[:4], 'little')
        assert l == len(data) - 4
        return data[4:]


class U32:
    def __init__(self, data: int) -> None:
        assert 0 <= data and data <= 0xffffffff
        self.data = data

    def __eq__(self, other) -> bool:
        return self.data == other.data

    def molecule(self) -> bytearray:
        return self.data.to_bytes(4, 'little')

    @classmethod
    def molecule_decode(cls, data: bytearray) -> int:
        return int.from_bytes(data, 'little')

    @classmethod
    def molecule_size(cls) -> int:
        return 4


class U64:
    def __init__(self, data: int) -> None:
        assert 0 <= data and data <= 0xffffffffffffffff
        self.data = data

    def __eq__(self, other) -> bool:
        return self.data == other.data

    def molecule(self) -> bytearray:
        return self.data.to_bytes(8, 'little')

    @classmethod
    def molecule_decode(cls, data: bytearray) -> int:
        return int.from_bytes(data, 'little')

    @classmethod
    def molecule_size(cls) -> int:
        return 8


class U128:
    def __init__(self, data: int) -> None:
        assert 0 <= data and data <= 0xffffffffffffffffffffffffffffffff
        self.data = data

    def __eq__(self, other) -> bool:
        return self.data == other.data

    def molecule(self) -> bytearray:
        return self.data.to_bytes(16, 'little')

    @classmethod
    def molecule_decode(cls, data: bytearray) -> int:
        return int.from_bytes(data, 'little')

    @classmethod
    def molecule_size(cls) -> int:
        return 16



import random
import requests
import typing


class ObjectDict(dict):
    def __getattr__(self, name: str) -> typing.Any:
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, name: str, value: typing.Any):
        self[name] = value


develop = ObjectDict({
    'hrp': 'ckt',
    'url': 'http://127.0.0.1:8114',
    'script': ObjectDict({
        'dao': ObjectDict({
            'code_hash': bytearray.fromhex('82d76d1b75fe2fd9a27dfbaa65a039221a380d76c926f378d3f81cf3e7e13f2e'),
            'hash_type': 1,
            'cell_dep': ObjectDict({
                'out_point': ObjectDict({
                    'tx_hash': bytearray.fromhex('0000000000000000000000000000000000000000000000000000000000000000'),
                    'index': 2,
                }),
                'dep_type': 0,
            })
        }),
        'secp256k1_blake160': ObjectDict({
            'code_hash': bytearray.fromhex('9bd7e06f3ecf4be0f2fcd2188b23f1b9fcc88e5d4b65a8637b17723bbda3cce8'),
            'hash_type': 1,
            'cell_dep': ObjectDict({
                'out_point': ObjectDict({
                    'tx_hash': bytearray.fromhex('0000000000000000000000000000000000000000000000000000000000000000'),
                    'index': 0,
                }),
                'dep_type': 1,
            })
        }),
    })
})

mainnet = ObjectDict({
    'hrp': 'ckb',
    # https://github.com/nervosnetwork/ckb/wiki/Public-JSON-RPC-nodes
    'url': 'https://mainnet.ckbapp.dev',
    'script': ObjectDict({
        'dao': ObjectDict({
            'code_hash': bytearray.fromhex('82d76d1b75fe2fd9a27dfbaa65a039221a380d76c926f378d3f81cf3e7e13f2e'),
            'hash_type': 1,
            'cell_dep': ObjectDict({
                'out_point': ObjectDict({
                    'tx_hash': bytearray.fromhex('e2fb199810d49a4d8beec56718ba2593b665db9d52299a0f9e6e75416d73ff5c'),
                    'index': 2,
                }),
                'dep_type': 0,
            })
        }),
        'secp256k1_blake160': ObjectDict({
            'code_hash': bytearray.fromhex('9bd7e06f3ecf4be0f2fcd2188b23f1b9fcc88e5d4b65a8637b17723bbda3cce8'),
            'hash_type': 1,
            'cell_dep': ObjectDict({
                'out_point': ObjectDict({
                    'tx_hash': bytearray.fromhex('71a7ba8fc96349fea0ed3a5c47992e3b4084b031a42264a018e0072e8172e46c'),
                    'index': 0,
                }),
                'dep_type': 1,
            })
        }),
    })
})

testnet = ObjectDict({
    'hrp': 'ckt',
    # https://github.com/nervosnetwork/ckb/wiki/Public-JSON-RPC-nodes
    'url': 'https://testnet.ckbapp.dev',
    'script': ObjectDict({
        'dao': ObjectDict({
            'code_hash': bytearray.fromhex('82d76d1b75fe2fd9a27dfbaa65a039221a380d76c926f378d3f81cf3e7e13f2e'),
            'hash_type': 1,
            'cell_dep': ObjectDict({
                'out_point': ObjectDict({
                    'tx_hash': bytearray.fromhex('8f8c79eb6671709633fe6a46de93c0fedc9c1b8a6527a18d3983879542635c9f'),
                    'index': 2,
                }),
                'dep_type': 0,
            })
        }),
        'secp256k1_blake160': ObjectDict({
            'code_hash': bytearray.fromhex('9bd7e06f3ecf4be0f2fcd2188b23f1b9fcc88e5d4b65a8637b17723bbda3cce8'),
            'hash_type': 1,
            'cell_dep': ObjectDict({
                'out_point': ObjectDict({
                    'tx_hash': bytearray.fromhex('f8de3bb47d055cdf460d93a2a6e1b05f7432f9777c8c474abf4eec1d4aee5d37'),
                    'index': 0,
                }),
                'dep_type': 1,
            })
        }),
    })
})


def upgrade(url: str):
    r = requests.post(url, json={
        'id': random.randint(0x00000000, 0xffffffff),
        'jsonrpc': '2.0',
        'method': 'get_block_by_number',
        'params': ['0x0']
    })
    t = r.json()['result']['transactions']
    develop.url = url
    develop.script.dao.cell_dep.out_point.tx_hash = bytearray.fromhex(t[0]['hash'][2:])
    develop.script.secp256k1_blake160.cell_dep.out_point.tx_hash = bytearray.fromhex(t[1]['hash'][2:])


current = develop



import pytest
import subprocess


def call(c: str):
    return subprocess.run(c, check=True, shell=True)


def test_addr():
    call('python example/addr.py --prikey 1')


def test_capacity():
    call('python example/capacity.py --addr ckt1qzda0cr08m85hc8jlnfp3zer7xulejywt49kt2rr0vthywaa50xwsqt4z78ng4yutl5u6xsv27ht6q08mhujf8s2r0n40')


def test_deploy():
    call('python example/deploy.py --prikey 1 --file LICENSE')


@pytest.mark.skip()
def test_faucet():
    call('python example/faucet.py --addr ckt1qzda0cr08m85hc8jlnfp3zer7xulejywt49kt2rr0vthywaa50xwsqt4z78ng4yutl5u6xsv27ht6q08mhujf8s2r0n40')


def test_redeem():
    call('python example/redeem.py --prikey 1')


def test_transfer():
    call('python example/transfer.py --prikey 1 --to ckt1qzda0cr08m85hc8jlnfp3zer7xulejywt49kt2rr0vthywaa50xwsqdrcaufs8qeu8wvvy0myyedek4vqad9qeq3gc4cf --value 100')


def test_txdump():
    call('python example/txdump.py --net testnet --hash 0x123b09a89e65cc9c375dab739c9c921f7067d0b205e563135bb5a1221f8948d9')



import itertools
import pyckb.config
import pyckb.core
import random
import requests
import time
import typing

# Doc: https://github.com/nervosnetwork/ckb/tree/develop/rpc


def call(method: str, params: typing.List) -> typing.Any:
    r = requests.post(pyckb.config.current.url, json={
        'id': random.randint(0x00000000, 0xffffffff),
        'jsonrpc': '2.0',
        'method': method,
        'params': params,
    }).json()
    if 'error' in r:
        raise Exception(r['error'])
    return r['result']


def wait(hash: str):
    for _ in itertools.repeat(0):
        time.sleep(1)
        r = get_transaction(hash)
        if r['tx_status']['status'] == 'committed':
            break


def add_node():
    pass


def calculate_dao_field():
    pass


def calculate_dao_maximum_withdraw():
    pass


def clear_banned_addresses():
    pass


def clear_tx_pool():
    pass


def estimate_cycles():
    pass


def generate_block():
    pass


def generate_block_with_template():
    pass


def generate_epochs():
    pass


def get_banned_addresses():
    pass


def get_block(block_hash: str) -> typing.Dict:
    return call('get_block', [block_hash])


def get_block_by_number(block_number: str) -> typing.Dict:
    return call('get_block_by_number', [block_number])


def get_block_economic_state():
    pass


def get_block_filter():
    pass


def get_block_hash():
    pass


def get_block_median_time():
    pass


def get_block_template():
    pass


def get_blockchain_info():
    pass


def get_cells(search_key: typing.Dict, order: str, limit: str, after: str) -> typing.Dict:
    return call('get_cells', [search_key, order, limit, after])


def get_cells_capacity(search_key: typing.Dict) -> typing.Dict:
    return call('get_cells_capacity', [search_key])


def get_cells_iter(search_key: typing.Dict) -> typing.Generator:
    cursor = None
    limits = 256
    for _ in itertools.repeat(0):
        r = get_cells(search_key, 'asc', hex(limits), cursor)
        cursor = r['last_cursor']
        for e in r['objects']:
            yield e
        if len(r['objects']) < limits:
            break


def get_consensus():
    pass


def get_current_epoch() -> typing.Dict[str, typing.Any]:
    return call('get_current_epoch', [])


def get_deployments_info():
    pass


def get_epoch_by_number():
    pass


def get_fee_rate_statistics():
    pass


def get_fork_block():
    pass


def get_header(block_hash: str) -> typing.Dict:
    return call('get_header', [block_hash])


def get_header_by_number(block_number: str) -> typing.Dict:
    return call('get_header_by_number', [block_number])


def get_indexer_tip() -> typing.Dict:
    return call('get_indexer_tip', [])


def get_live_cell():
    pass


def get_peers():
    pass


def get_pool_tx_detail_info():
    pass


def get_raw_tx_pool():
    pass


def get_tip_block_number() -> str:
    return call('get_tip_block_number', [])


def get_tip_header() -> typing.Dict:
    return call('get_tip_header', [])


def get_transaction(tx_hash: str) -> typing.Dict:
    return call('get_transaction', [tx_hash])


def get_transaction_and_witness_proof():
    pass


def get_transaction_proof():
    pass


def get_transactions():
    pass


def jemalloc_profiling_dump():
    pass


def local_node_info():
    pass


def notify_transaction():
    pass


def ping_peers():
    pass


def process_block_without_verify():
    pass


def remove_node():
    pass


def remove_transaction():
    pass


def send_alert():
    pass


def send_transaction(transaction: typing.Dict) -> str:
    return call('send_transaction', [transaction, 'passthrough'])


def set_ban():
    pass


def set_extra_logger():
    pass


def set_network_active():
    pass


def submit_block():
    pass


def subscribe():
    pass


def sync_state():
    pass


def test_tx_pool_accept():
    pass


def truncate():
    pass


def tx_pool_info():
    pass


def tx_pool_ready():
    pass


def unsubscribe():
    pass


def update_main_logger():
    pass


def verify_transaction_and_witness_proof():
    pass


def verify_transaction_proof():
    pass



import itertools
import pyckb.secp256k1
import random
import typing


def sign(prikey: pyckb.secp256k1.Fr, m: pyckb.secp256k1.Fr) -> typing.Tuple[pyckb.secp256k1.Fr, pyckb.secp256k1.Fr, int]:
    # https://www.secg.org/sec1-v2.pdf
    # 4.1.3 Signing Operation
    for _ in itertools.repeat(0):
        k = pyckb.secp256k1.Fr(random.randint(0, pyckb.secp256k1.N - 1))
        R = pyckb.secp256k1.G * k
        r = pyckb.secp256k1.Fr(R.x.x)
        if r.x == 0:
            continue
        s = (m + prikey * r) / k
        if s.x == 0:
            continue
        v = 0
        if R.y.x & 1 == 1:
            v |= 1
        if R.x.x >= pyckb.secp256k1.N:
            v |= 2
        return r, s, v


def verify(pubkey: pyckb.secp256k1.Pt, m: pyckb.secp256k1.Fr, r: pyckb.secp256k1.Fr, s: pyckb.secp256k1.Fr) -> bool:
    # https://www.secg.org/sec1-v2.pdf
    # 4.1.4 Verifying Operation
    u1 = m / s
    u2 = r / s
    x = pyckb.secp256k1.G * u1 + pubkey * u2
    assert x != pyckb.secp256k1.I
    v = pyckb.secp256k1.Fr(x.x.x)
    return v == r


def pubkey(m: pyckb.secp256k1.Fr, r: pyckb.secp256k1.Fr, s: pyckb.secp256k1.Fr, v: int) -> pyckb.secp256k1.Pt:
    # https://www.secg.org/sec1-v2.pdf
    # 4.1.6 Public Key Recovery Operation
    assert v in [0, 1, 2, 3]
    if v & 2 == 0:
        x = pyckb.secp256k1.Fq(r.x)
    else:
        x = pyckb.secp256k1.Fq(r.x + pyckb.secp256k1.N)
    y_y = x * x * x + pyckb.secp256k1.A * x + pyckb.secp256k1.B
    y = y_y ** ((pyckb.secp256k1.P + 1) // 4)
    if v & 1 != y.x & 1:
        y = -y
    R = pyckb.secp256k1.Pt(x, y)
    return (R * s - pyckb.secp256k1.G * m) / r



import pyckb


def test_wallet_addr():
    pyckb.config.upgrade('http://127.0.0.1:8114')
    pyckb.config.current = pyckb.config.develop
    user = pyckb.wallet.Wallet(1)
    addr = user.addr
    assert addr == 'ckt1qzda0cr08m85hc8jlnfp3zer7xulejywt49kt2rr0vthywaa50xwsqt4z78ng4yutl5u6xsv27ht6q08mhujf8s2r0n40'


def test_wallet_transfer():
    pyckb.config.upgrade('http://127.0.0.1:8114')
    pyckb.config.current = pyckb.config.develop
    user = pyckb.wallet.Wallet(1)
    mate = pyckb.wallet.Wallet(2)
    capacity = 100 * pyckb.denomination.ckbytes
    capacity_old = mate.capacity()
    hash = user.transfer(mate.script, capacity)
    pyckb.rpc.wait(f'0x{hash.hex()}')
    capacity_new = mate.capacity()
    assert capacity_new - capacity_old == capacity
    hash = mate.transfer_all(user.script)
    pyckb.rpc.wait(f'0x{hash.hex()}')
    assert mate.capacity() == 0


def test_wallet_script_deploy():
    pyckb.config.upgrade('http://127.0.0.1:8114')
    pyckb.config.current = pyckb.config.develop
    user = pyckb.wallet.Wallet(1)
    hash = user.script_deploy(user.script, bytearray([0, 1, 2, 3]))
    pyckb.rpc.wait(f'0x{hash.hex()}')


def test_wallet_script_deploy_type_id():
    pyckb.config.upgrade('http://127.0.0.1:8114')
    pyckb.config.current = pyckb.config.develop
    user = pyckb.wallet.Wallet(1)
    hash = user.script_deploy_type_id(user.script, bytearray([0, 1, 2, 3]))
    pyckb.rpc.wait(f'0x{hash.hex()}')
    out_point = pyckb.core.OutPoint(hash, 0)
    hash = user.script_update_type_id(user.script, bytearray([0, 1, 2, 3, 4, 5]), out_point)
    pyckb.rpc.wait(f'0x{hash.hex()}')


def test_wallet_dao():
    pyckb.config.upgrade('http://127.0.0.1:8114')
    pyckb.config.current = pyckb.config.develop
    user = pyckb.wallet.Wallet(1)
    hash = user.dao_deposit(200 * pyckb.denomination.ckbytes)
    pyckb.rpc.wait(f'0x{hash.hex()}')
    hash = user.dao_prepare(pyckb.core.OutPoint(hash, 0))
    pyckb.rpc.wait(f'0x{hash.hex()}')



import itertools
import json
import math
import pyckb.config
import pyckb.denomination
import pyckb.core
import pyckb.rpc
import typing


class WalletTransactionAnalyzer:
    def __init__(self, tx: pyckb.core.Transaction) -> None:
        self.tx = tx

    def analyze_mining_fee(self) -> None:
        # Make sure the transaction fee is less than 1 CKB. This is a rough check, but works well in most cases.
        sender_capacity = 0
        output_capacity = 0
        for e in self.tx.raw.inputs:
            out_point = e.previous_output
            result = pyckb.rpc.get_transaction('0x' + out_point.tx_hash.hex())
            origin = pyckb.core.CellOutput.json_decode(result['transaction']['outputs'][out_point.index])
            sender_capacity += origin.capacity
        for e in self.tx.raw.outputs:
            output_capacity += e.capacity
        assert sender_capacity - output_capacity <= 1 * pyckb.denomination.ckbytes

    def analyze_outputs_data(self) -> None:
        assert len(self.tx.raw.outputs) == len(self.tx.raw.outputs_data)

    def analyze_outputs_lock(self) -> None:
        for e in self.tx.raw.outputs:
            if e.lock.code_hash == pyckb.config.current.script.secp256k1_blake160.code_hash:
                assert e.lock.hash_type == pyckb.config.current.script.secp256k1_blake160.hash_type
                assert len(e.lock.args) == 20

    def analyze_since(self) -> None:
        # Transaction since precondition
        # See https://github.com/nervosnetwork/rfcs/blob/master/rfcs/0017-tx-valid-since/0017-tx-valid-since.md
        for e in self.tx.raw.inputs:
            if e.since == 0:
                continue
            if e.since >> 56 == 0x00:
                pass
            if e.since >> 56 == 0x20:
                current_epoch = pyckb.core.epoch_decode(int(pyckb.rpc.get_tip_header()['epoch'], 16))
                request_epoch = pyckb.core.epoch_decode(e.since & 0xffffffffffffff)
                if current_epoch[0] == request_epoch[0]:
                    assert current_epoch[1] >= request_epoch[1]
                else:
                    assert current_epoch[0] >= request_epoch[0]
            if e.since >> 56 == 0x60:
                pass
            if e.since >> 56 == 0x80:
                pass
            if e.since >> 56 == 0xa0:
                pass
            if e.since >> 56 == 0xe0:
                pass

    def analyze(self) -> None:
        self.analyze_mining_fee()
        self.analyze_outputs_data()
        self.analyze_outputs_lock()
        self.analyze_since()


class Wallet:
    def __init__(self, prikey: int) -> None:
        self.prikey = pyckb.core.PriKey(prikey)
        self.pubkey = self.prikey.pubkey()
        self.script = pyckb.core.Script(
            pyckb.config.current.script.secp256k1_blake160.code_hash,
            pyckb.config.current.script.secp256k1_blake160.hash_type,
            pyckb.core.hash(self.pubkey.sec())[:20]
        )
        self.addr = pyckb.core.address_encode(self.script)

    def __repr__(self) -> str:
        return json.dumps(self.json())

    def __eq__(self, other) -> bool:
        return all([
            self.prikey == other.prikey,
            self.pubkey == other.pubkey,
            self.script == other.script,
            self.addr == other.addr,
        ])

    def json(self) -> typing.Dict:
        return {
            'prikey': self.prikey.json(),
            'pubkey': self.pubkey.json(),
            'script': self.script.json(),
            'addr': self.addr,
        }

    def livecell(self) -> typing.Generator:
        return pyckb.rpc.get_cells_iter({
            'script': self.script.json(),
            'script_type': 'lock',
            'filter': {
                'script_len_range': ['0x0', '0x1']
            }
        })

    def capacity(self) -> int:
        return int(pyckb.rpc.get_cells_capacity({
            'script': self.script.json(),
            'script_type': 'lock',
            'filter': {
                'script_len_range': ['0x0', '0x1']
            }
        })['capacity'], 16)

    def transfer(self, script: pyckb.core.Script, capacity: int) -> bytearray:
        assert capacity >= 61 * pyckb.denomination.ckbytes
        assert self.capacity() > capacity
        sender_capacity = 0
        accept_capacity = capacity
        accept_script = script
        change_capacity = 0
        change_script = self.script
        tx = pyckb.core.Transaction(pyckb.core.RawTransaction(0, [], [], [], [], []), [])
        tx.raw.cell_deps.append(pyckb.core.CellDep.conf_decode(pyckb.config.current.script.secp256k1_blake160.cell_dep))
        tx.raw.outputs.append(pyckb.core.CellOutput(accept_capacity, accept_script, None))
        tx.raw.outputs.append(pyckb.core.CellOutput(change_capacity, change_script, None))
        tx.raw.outputs_data.append(bytearray())
        tx.raw.outputs_data.append(bytearray())
        tx.witnesses.append(pyckb.core.WitnessArgs(bytearray(65), None, None).molecule())
        for cell in itertools.islice(self.livecell(), 256):
            cell_out_point = pyckb.core.OutPoint.json_decode(cell['out_point'])
            cell_capacity = int(cell['output']['capacity'], 16)
            cell_input = pyckb.core.CellInput(0, cell_out_point)
            sender_capacity += cell_capacity
            tx.raw.inputs.append(cell_input)
            change_capacity = sender_capacity - accept_capacity - len(tx.molecule()) - 4
            if change_capacity >= 61 * pyckb.denomination.ckbytes:
                break
        assert change_capacity >= 61 * pyckb.denomination.ckbytes
        tx.raw.outputs[1].capacity = change_capacity
        sign_data = bytearray()
        sign_data.extend(tx.raw.hash())
        sign_data.extend(len(tx.witnesses[0]).to_bytes(8, 'little'))
        sign_data.extend(tx.witnesses[0])
        sign_data = pyckb.core.hash(sign_data)
        sign = self.prikey.sign(sign_data)
        tx.witnesses[0] = pyckb.core.WitnessArgs(sign, None, None).molecule()
        WalletTransactionAnalyzer(tx).analyze()
        hash = pyckb.rpc.send_transaction(tx.json())
        return bytearray.fromhex(hash[2:])

    def transfer_all(self, script: pyckb.core.Script) -> bytearray:
        assert self.capacity() > 0
        sender_capacity = 0
        accept_capacity = 0
        accept_script = script
        tx = pyckb.core.Transaction(pyckb.core.RawTransaction(0, [], [], [], [], []), [])
        tx.raw.cell_deps.append(pyckb.core.CellDep.conf_decode(pyckb.config.current.script.secp256k1_blake160.cell_dep))
        tx.raw.outputs.append(pyckb.core.CellOutput(accept_capacity, accept_script, None))
        tx.raw.outputs_data.append(bytearray())
        tx.witnesses.append(pyckb.core.WitnessArgs(bytearray(65), None, None).molecule())
        for cell in itertools.islice(self.livecell(), 256):
            cell_out_point = pyckb.core.OutPoint.json_decode(cell['out_point'])
            cell_capacity = int(cell['output']['capacity'], 16)
            cell_input = pyckb.core.CellInput(0, cell_out_point)
            sender_capacity += cell_capacity
            tx.raw.inputs.append(cell_input)
        accept_capacity = sender_capacity - len(tx.molecule()) - 4
        tx.raw.outputs[0].capacity = accept_capacity
        sign_data = bytearray()
        sign_data.extend(tx.raw.hash())
        sign_data.extend(len(tx.witnesses[0]).to_bytes(8, 'little'))
        sign_data.extend(tx.witnesses[0])
        sign_data = pyckb.core.hash(sign_data)
        sign = self.prikey.sign(sign_data)
        tx.witnesses[0] = pyckb.core.WitnessArgs(sign, None, None).molecule()
        WalletTransactionAnalyzer(tx).analyze()
        hash = pyckb.rpc.send_transaction(tx.json())
        return bytearray.fromhex(hash[2:])

    def script_deploy(self, script: pyckb.core.Script, data: bytearray) -> bytearray:
        sender_capacity = 0
        accept_capacity = (61 + len(data)) * pyckb.denomination.ckbytes
        accept_script = script
        change_capacity = 0
        change_script = self.script
        tx = pyckb.core.Transaction(pyckb.core.RawTransaction(0, [], [], [], [], []), [])
        tx.raw.cell_deps.append(pyckb.core.CellDep.conf_decode(pyckb.config.current.script.secp256k1_blake160.cell_dep))
        tx.raw.outputs.append(pyckb.core.CellOutput(accept_capacity, accept_script, None))
        tx.raw.outputs.append(pyckb.core.CellOutput(change_capacity, change_script, None))
        tx.raw.outputs_data.append(data)
        tx.raw.outputs_data.append(bytearray())
        tx.witnesses.append(pyckb.core.WitnessArgs(bytearray(65), None, None).molecule())
        for cell in itertools.islice(self.livecell(), 256):
            cell_out_point = pyckb.core.OutPoint.json_decode(cell['out_point'])
            cell_capacity = int(cell['output']['capacity'], 16)
            cell_input = pyckb.core.CellInput(0, cell_out_point)
            sender_capacity += cell_capacity
            tx.raw.inputs.append(cell_input)
            change_capacity = sender_capacity - accept_capacity - len(tx.molecule()) - 4
            if change_capacity >= 61 * pyckb.denomination.ckbytes:
                break
        assert change_capacity >= 61 * pyckb.denomination.ckbytes
        tx.raw.outputs[1].capacity = change_capacity
        sign_data = bytearray()
        sign_data.extend(tx.raw.hash())
        sign_data.extend(len(tx.witnesses[0]).to_bytes(8, 'little'))
        sign_data.extend(tx.witnesses[0])
        sign_data = pyckb.core.hash(sign_data)
        sign = self.prikey.sign(sign_data)
        tx.witnesses[0] = pyckb.core.WitnessArgs(sign, None, None).molecule()
        WalletTransactionAnalyzer(tx).analyze()
        hash = pyckb.rpc.send_transaction(tx.json())
        return bytearray.fromhex(hash[2:])

    def script_deploy_type_id(self, script: pyckb.core.Script, data: bytearray) -> bytearray:
        sender_capacity = 0
        accept_capacity = (126 + len(data)) * pyckb.denomination.ckbytes
        accept_script = script
        accept_typeid = pyckb.core.Script(pyckb.core.type_id_code_hash, pyckb.core.type_id_hash_type, bytearray(32))
        change_capacity = 0
        change_script = self.script
        tx = pyckb.core.Transaction(pyckb.core.RawTransaction(0, [], [], [], [], []), [])
        tx.raw.cell_deps.append(pyckb.core.CellDep.conf_decode(pyckb.config.current.script.secp256k1_blake160.cell_dep))
        tx.raw.outputs.append(pyckb.core.CellOutput(accept_capacity, accept_script, accept_typeid))
        tx.raw.outputs.append(pyckb.core.CellOutput(change_capacity, change_script, None))
        tx.raw.outputs_data.append(data)
        tx.raw.outputs_data.append(bytearray())
        tx.witnesses.append(pyckb.core.WitnessArgs(bytearray(65), None, None).molecule())
        for cell in itertools.islice(self.livecell(), 256):
            cell_out_point = pyckb.core.OutPoint.json_decode(cell['out_point'])
            cell_capacity = int(cell['output']['capacity'], 16)
            cell_input = pyckb.core.CellInput(0, cell_out_point)
            sender_capacity += cell_capacity
            tx.raw.inputs.append(cell_input)
            change_capacity = sender_capacity - accept_capacity - len(tx.molecule()) - 4
            if change_capacity >= 61 * pyckb.denomination.ckbytes:
                break
        assert change_capacity >= 61 * pyckb.denomination.ckbytes
        # https://github.com/nervosnetwork/rfcs/blob/master/rfcs/0022-transaction-structure/0022-transaction-structure.md#type-id
        tx.raw.outputs[0].type.args = pyckb.core.hash(tx.raw.inputs[0].molecule() + bytearray(8))
        tx.raw.outputs[1].capacity = change_capacity
        sign_data = bytearray()
        sign_data.extend(tx.raw.hash())
        sign_data.extend(len(tx.witnesses[0]).to_bytes(8, 'little'))
        sign_data.extend(tx.witnesses[0])
        sign_data = pyckb.core.hash(sign_data)
        sign = self.prikey.sign(sign_data)
        tx.witnesses[0] = pyckb.core.WitnessArgs(sign, None, None).molecule()
        WalletTransactionAnalyzer(tx).analyze()
        hash = pyckb.rpc.send_transaction(tx.json())
        return bytearray.fromhex(hash[2:])

    def script_update_type_id(
        self,
        script: pyckb.core.Script,
        data: bytearray,
        out_point: pyckb.core.OutPoint
    ) -> bytearray:
        result = pyckb.rpc.get_transaction('0x' + out_point.tx_hash.hex())
        origin = pyckb.core.CellOutput.json_decode(result['transaction']['outputs'][out_point.index])
        assert origin.type.code_hash == pyckb.core.type_id_code_hash
        assert origin.type.hash_type == pyckb.core.type_id_hash_type
        sender_capacity = origin.capacity
        accept_capacity = (126 + len(data)) * pyckb.denomination.ckbytes
        accept_script = script
        accept_typeid = origin.type
        change_capacity = 0
        change_script = self.script
        tx = pyckb.core.Transaction(pyckb.core.RawTransaction(0, [], [], [], [], []), [])
        tx.raw.cell_deps.append(pyckb.core.CellDep.conf_decode(pyckb.config.current.script.secp256k1_blake160.cell_dep))
        tx.raw.inputs.append(pyckb.core.CellInput(0, out_point))
        tx.raw.outputs.append(pyckb.core.CellOutput(accept_capacity, accept_script, accept_typeid))
        tx.raw.outputs.append(pyckb.core.CellOutput(change_capacity, change_script, None))
        tx.raw.outputs_data.append(data)
        tx.raw.outputs_data.append(bytearray())
        tx.witnesses.append(pyckb.core.WitnessArgs(bytearray(65), None, None).molecule())
        for cell in itertools.islice(self.livecell(), 255):
            cell_out_point = pyckb.core.OutPoint.json_decode(cell['out_point'])
            cell_capacity = int(cell['output']['capacity'], 16)
            cell_input = pyckb.core.CellInput(0, cell_out_point)
            sender_capacity += cell_capacity
            tx.raw.inputs.append(cell_input)
            change_capacity = sender_capacity - accept_capacity - len(tx.molecule()) - 4
            if change_capacity >= 61 * pyckb.denomination.ckbytes:
                break
        assert change_capacity >= 61 * pyckb.denomination.ckbytes
        tx.raw.outputs[1].capacity = change_capacity
        sign_data = bytearray()
        sign_data.extend(tx.raw.hash())
        sign_data.extend(len(tx.witnesses[0]).to_bytes(8, 'little'))
        sign_data.extend(tx.witnesses[0])
        sign_data = pyckb.core.hash(sign_data)
        sign = self.prikey.sign(sign_data)
        tx.witnesses[0] = pyckb.core.WitnessArgs(sign, None, None).molecule()
        WalletTransactionAnalyzer(tx).analyze()
        hash = pyckb.rpc.send_transaction(tx.json())
        return bytearray.fromhex(hash[2:])

    def dao_deposit(self, capacity: int) -> bytearray:
        # https://github.com/nervosnetwork/rfcs/blob/master/rfcs/0023-dao-deposit-withdraw/0023-dao-deposit-withdraw.md#deposit
        assert capacity >= 102 * pyckb.denomination.ckbytes
        assert self.capacity() > capacity
        sender_capacity = 0
        accept_capacity = capacity
        accept_script = self.script
        accept_typeid = pyckb.core.Script(
            pyckb.config.current.script.dao.code_hash,
            pyckb.config.current.script.dao.hash_type,
            bytearray()
        )
        change_capacity = 0
        change_script = self.script
        tx = pyckb.core.Transaction(pyckb.core.RawTransaction(0, [], [], [], [], []), [])
        tx.raw.cell_deps.append(pyckb.core.CellDep.conf_decode(pyckb.config.current.script.secp256k1_blake160.cell_dep))
        tx.raw.cell_deps.append(pyckb.core.CellDep.conf_decode(pyckb.config.current.script.dao.cell_dep))
        tx.raw.outputs.append(pyckb.core.CellOutput(accept_capacity, accept_script, accept_typeid))
        tx.raw.outputs.append(pyckb.core.CellOutput(change_capacity, change_script, None))
        tx.raw.outputs_data.append(bytearray(8))
        tx.raw.outputs_data.append(bytearray())
        tx.witnesses.append(pyckb.core.WitnessArgs(bytearray(65), None, None).molecule())
        for cell in itertools.islice(self.livecell(), 256):
            cell_out_point = pyckb.core.OutPoint.json_decode(cell['out_point'])
            cell_capacity = int(cell['output']['capacity'], 16)
            cell_input = pyckb.core.CellInput(0, cell_out_point)
            sender_capacity += cell_capacity
            tx.raw.inputs.append(cell_input)
            change_capacity = sender_capacity - accept_capacity - len(tx.molecule()) - 4
            if change_capacity >= 61 * pyckb.denomination.ckbytes:
                break
        assert change_capacity >= 61 * pyckb.denomination.ckbytes
        tx.raw.outputs[1].capacity = change_capacity
        sign_data = bytearray()
        sign_data.extend(tx.raw.hash())
        sign_data.extend(len(tx.witnesses[0]).to_bytes(8, 'little'))
        sign_data.extend(tx.witnesses[0])
        sign_data = pyckb.core.hash(sign_data)
        sign = self.prikey.sign(sign_data)
        tx.witnesses[0] = pyckb.core.WitnessArgs(sign, None, None).molecule()
        WalletTransactionAnalyzer(tx).analyze()
        hash = pyckb.rpc.send_transaction(tx.json())
        return bytearray.fromhex(hash[2:])

    def dao_prepare(self, out_point: pyckb.core.OutPoint) -> bytearray:
        # https://github.com/nervosnetwork/rfcs/blob/master/rfcs/0023-dao-deposit-withdraw/0023-dao-deposit-withdraw.md#withdraw-phase-1
        result = pyckb.rpc.get_transaction('0x' + out_point.tx_hash.hex())
        number = int(pyckb.rpc.get_header(result['tx_status']['block_hash'])['number'], 16)
        origin = pyckb.core.CellOutput.json_decode(result['transaction']['outputs'][out_point.index])
        assert origin.type.code_hash == pyckb.config.current.script.dao.code_hash
        assert origin.type.hash_type == pyckb.config.current.script.dao.hash_type
        assert origin.type.args == bytearray()
        sender_capacity = origin.capacity
        accept_capacity = origin.capacity
        accept_script = origin.lock
        accept_typeid = origin.type
        change_capacity = 0
        change_script = self.script
        tx = pyckb.core.Transaction(pyckb.core.RawTransaction(0, [], [], [], [], []), [])
        tx.raw.cell_deps.append(pyckb.core.CellDep.conf_decode(pyckb.config.current.script.secp256k1_blake160.cell_dep))
        tx.raw.cell_deps.append(pyckb.core.CellDep.conf_decode(pyckb.config.current.script.dao.cell_dep))
        tx.raw.header_deps.append(bytearray.fromhex(result['tx_status']['block_hash'][2:]))
        tx.raw.inputs.append(pyckb.core.CellInput(0, out_point))
        tx.raw.outputs.append(pyckb.core.CellOutput(accept_capacity, accept_script, accept_typeid))
        tx.raw.outputs.append(pyckb.core.CellOutput(change_capacity, change_script, None))
        tx.raw.outputs_data.append(number.to_bytes(8, 'little'))
        tx.raw.outputs_data.append(bytearray())
        tx.witnesses.append(pyckb.core.WitnessArgs(bytearray(65), None, None).molecule())
        for cell in itertools.islice(self.livecell(), 255):
            cell_out_point = pyckb.core.OutPoint.json_decode(cell['out_point'])
            cell_capacity = int(cell['output']['capacity'], 16)
            cell_input = pyckb.core.CellInput(0, cell_out_point)
            sender_capacity += cell_capacity
            tx.raw.inputs.append(cell_input)
            change_capacity = sender_capacity - accept_capacity - len(tx.molecule()) - 4
            if change_capacity >= 61 * pyckb.denomination.ckbytes:
                break
        assert change_capacity >= 61 * pyckb.denomination.ckbytes
        tx.raw.outputs[1].capacity = change_capacity
        sign_data = bytearray()
        sign_data.extend(tx.raw.hash())
        sign_data.extend(len(tx.witnesses[0]).to_bytes(8, 'little'))
        sign_data.extend(tx.witnesses[0])
        sign_data = pyckb.core.hash(sign_data)
        sign = self.prikey.sign(sign_data)
        tx.witnesses[0] = pyckb.core.WitnessArgs(sign, None, None).molecule()
        WalletTransactionAnalyzer(tx).analyze()
        hash = pyckb.rpc.send_transaction(tx.json())
        return bytearray.fromhex(hash[2:])

    def dao_extract(self, out_point: pyckb.core.OutPoint) -> bytearray:
        # https://github.com/nervosnetwork/rfcs/blob/master/rfcs/0023-dao-deposit-withdraw/0023-dao-deposit-withdraw.md#withdraw-phase-2
        result = pyckb.rpc.get_transaction('0x' + out_point.tx_hash.hex())
        origin = pyckb.core.CellOutput.json_decode(result['transaction']['outputs'][out_point.index])
        assert origin.type.code_hash == pyckb.config.current.script.dao.code_hash
        assert origin.type.hash_type == pyckb.config.current.script.dao.hash_type
        assert origin.type.args == bytearray()
        deposit_block_number_byte = bytearray.fromhex(result['transaction']['outputs_data'][out_point.index][2:])
        deposit_block_number = int.from_bytes(deposit_block_number_byte, 'little')
        deposit_block_header = pyckb.core.Header.json_decode(pyckb.rpc.get_header_by_number(hex(deposit_block_number)))
        deposit_block_hash = deposit_block_header.hash()
        deposit_block_epoch = pyckb.core.epoch_decode(deposit_block_header.raw.epoch)
        deposit_block_epoch_float = deposit_block_epoch[0] + deposit_block_epoch[1] / deposit_block_epoch[2]
        deposit_dao_ar = pyckb.core.dao_decode(deposit_block_header.raw.dao)[1]
        prepare_block_hash = bytearray.fromhex(result['tx_status']['block_hash'][2:])
        prepare_block_header = pyckb.core.Header.json_decode(pyckb.rpc.get_header('0x' + prepare_block_hash.hex()))
        prepare_block_epoch = pyckb.core.epoch_decode(prepare_block_header.raw.epoch)
        prepare_block_epoch_float = prepare_block_epoch[0] + prepare_block_epoch[1] / prepare_block_epoch[2]
        prepare_dao_ar = pyckb.core.dao_decode(prepare_block_header.raw.dao)[1]
        extract_since_delay = math.ceil((prepare_block_epoch_float - deposit_block_epoch_float) / 180) * 180
        extract_since_epoch = pyckb.core.epoch_encode(
            deposit_block_epoch[0] + extract_since_delay,
            deposit_block_epoch[1],
            deposit_block_epoch[2],
        )
        extract_since = 0x2000000000000000 + extract_since_epoch
        occupy_capacity = 102 * pyckb.denomination.ckbytes
        sender_capacity = (origin.capacity - occupy_capacity) * prepare_dao_ar // deposit_dao_ar + occupy_capacity
        accept_capacity = 0
        accept_script = self.script
        tx = pyckb.core.Transaction(pyckb.core.RawTransaction(0, [], [], [], [], []), [])
        tx.raw.cell_deps.append(pyckb.core.CellDep.conf_decode(pyckb.config.current.script.secp256k1_blake160.cell_dep))
        tx.raw.cell_deps.append(pyckb.core.CellDep.conf_decode(pyckb.config.current.script.dao.cell_dep))
        tx.raw.header_deps.append(deposit_block_hash)
        tx.raw.header_deps.append(prepare_block_hash)
        tx.raw.inputs.append(pyckb.core.CellInput(extract_since, out_point))
        tx.raw.outputs.append(pyckb.core.CellOutput(accept_capacity, accept_script, None))
        tx.raw.outputs_data.append(bytearray())
        tx.witnesses.append(pyckb.core.WitnessArgs(bytearray(65), bytearray(8), None).molecule())
        accept_capacity = sender_capacity - len(tx.molecule()) - 4
        tx.raw.outputs[0].capacity = accept_capacity
        sign_data = bytearray()
        sign_data.extend(tx.raw.hash())
        sign_data.extend(len(tx.witnesses[0]).to_bytes(8, 'little'))
        sign_data.extend(tx.witnesses[0])
        sign_data = pyckb.core.hash(sign_data)
        sign = self.prikey.sign(sign_data)
        tx.witnesses[0] = pyckb.core.WitnessArgs(sign, bytearray(8), None).molecule()
        WalletTransactionAnalyzer(tx).analyze()
        hash = pyckb.rpc.send_transaction(tx.json())
        return bytearray.fromhex(hash[2:])

    def dao_livecell(self) -> typing.Generator:
        return pyckb.rpc.get_cells_iter({
            'script': self.script.json(),
            'script_type': 'lock',
            'filter': {
                'script': pyckb.core.Script(
                    pyckb.config.current.script.dao.code_hash,
                    pyckb.config.current.script.dao.hash_type,
                    bytearray(),
                ).json(),
            },
        })

    def dao_capacity(self) -> int:
        return int(pyckb.rpc.get_cells_capacity({
            'script': self.script.json(),
            'script_type': 'lock',
            'filter': {
                'script': pyckb.core.Script(
                    pyckb.config.current.script.dao.code_hash,
                    pyckb.config.current.script.dao.hash_type,
                    bytearray(),
                ).json(),
            }
        })['capacity'], 16)



import pyckb
import random


def test_sign():
    prikey = pyckb.secp256k1.Fr(random.randint(0, pyckb.secp256k1.N - 1))
    pubkey = pyckb.secp256k1.G * prikey
    m = pyckb.secp256k1.Fr(random.randint(0, pyckb.secp256k1.N - 1))
    r, s, _ = pyckb.ecdsa.sign(prikey, m)
    assert pyckb.ecdsa.verify(pubkey, m, r, s)


def test_pubkey():
    prikey = pyckb.secp256k1.Fr(random.randint(0, pyckb.secp256k1.N - 1))
    pubkey = pyckb.secp256k1.G * prikey
    m = pyckb.secp256k1.Fr(random.randint(0, pyckb.secp256k1.N - 1))
    r, s, v = pyckb.ecdsa.sign(prikey, m)
    assert pyckb.ecdsa.pubkey(m, r, s, v) == pubkey



import typing


class Fp:
    # Galois field. In mathematics, a finite field or Galois field is a field that contains a finite number of elements.
    # As with any field, a finite field is a set on which the operations of multiplication, addition, subtraction and
    # division are defined and satisfy certain basic rules.
    #
    # https://www.cs.miami.edu/home/burt/learning/Csc609.142/ecdsa-cert.pdf
    # Don Johnson, Alfred Menezes and Scott Vanstone, The Elliptic Curve Digital Signature Algorithm (ECDSA)
    # 3.1 The Finite Field Fp

    p = 0

    def __init__(self, x: int) -> None:
        self.x = x % self.p

    def __repr__(self) -> str:
        return f'Fp(0x{self.x:064x})'

    def __eq__(self, data: typing.Self) -> bool:
        assert self.p == data.p
        return self.x == data.x

    def __add__(self, data: typing.Self) -> typing.Self:
        assert self.p == data.p
        return self.__class__((self.x + data.x) % self.p)

    def __sub__(self, data: typing.Self) -> typing.Self:
        assert self.p == data.p
        return self.__class__((self.x - data.x) % self.p)

    def __mul__(self, data: typing.Self) -> typing.Self:
        assert self.p == data.p
        return self.__class__((self.x * data.x) % self.p)

    def __truediv__(self, data: typing.Self) -> typing.Self:
        return self * data ** -1

    def __pow__(self, data: int) -> typing.Self:
        return self.__class__(pow(self.x, data, self.p))

    def __pos__(self) -> typing.Self:
        return self

    def __neg__(self) -> typing.Self:
        return self.__class__(self.p - self.x)

    @classmethod
    def nil(cls) -> typing.Self:
        return cls(0)

    @classmethod
    def one(cls) -> typing.Self:
        return cls(1)


if __name__ == '__main__':
    Fp.p = 23
    assert Fp(12) + Fp(20) == Fp(9)
    assert Fp(8) * Fp(9) == Fp(3)
    assert Fp(8) ** -1 == Fp(3)
    Fp.p = 0

# Prime of finite field.
P = 0xfffffffffffffffffffffffffffffffffffffffffffffffffffffffefffffc2f
# The order n of G.
N = 0xfffffffffffffffffffffffffffffffebaaedce6af48a03bbfd25e8cd0364141


class Fq(Fp):

    p = P

    def __repr__(self) -> str:
        return f'Fq(0x{self.x:064x})'


class Fr(Fp):

    p = N

    def __repr__(self) -> str:
        return f'Fr(0x{self.x:064x})'


A = Fq(0)
B = Fq(7)


class Pt:

    def __init__(self, x: Fq, y: Fq) -> None:
        if x != Fq(0) or y != Fq(0):
            assert y ** 2 == x ** 3 + A * x + B
        self.x = x
        self.y = y

    def __repr__(self) -> str:
        return f'Pt({self.x}, {self.y})'

    def __eq__(self, data: typing.Self) -> bool:
        return all([
            self.x == data.x,
            self.y == data.y,
        ])

    def __add__(self, data: typing.Self) -> typing.Self:
        # https://www.cs.miami.edu/home/burt/learning/Csc609.142/ecdsa-cert.pdf
        # Don Johnson, Alfred Menezes and Scott Vanstone, The Elliptic Curve Digital Signature Algorithm (ECDSA)
        # 4.1 Elliptic Curves Over Fp
        if self.x == Fq(0) and self.y == Fq(0):
            return data
        if data.x == Fq(0) and data.y == Fq(0):
            return self
        if self.x == data.x and self.y == -data.y:
            return I
        x1, x2 = self.x, data.x
        y1, y2 = self.y, data.y
        if self.y == data.y:
            s = (x1 * x1 + x1 * x1 + x1 * x1 + A) / (y1 + y1)
        else:
            s = (y2 - y1) / (x2 - x1)
        x3 = s * s - x1 - x2
        y3 = s * (x1 - x3) - y1
        return Pt(x3, y3)

    def __sub__(self, data: typing.Self) -> typing.Self:
        return self + data.__neg__()

    def __mul__(self, k: Fr) -> typing.Self:
        # Point multiplication: Double-and-add
        # https://en.wikipedia.org/wiki/Elliptic_curve_point_multiplication
        n = k.x
        result = I
        addend = self
        while n:
            b = n & 1
            if b == 1:
                result += addend
            addend = addend + addend
            n = n >> 1
        return result

    def __truediv__(self, k: Fr) -> typing.Self:
        return self.__mul__(k ** -1)

    def __pos__(self) -> typing.Self:
        return self

    def __neg__(self) -> typing.Self:
        return Pt(self.x, -self.y)


# Identity element
I = Pt(
    Fq(0),
    Fq(0),
)
# Generator point
G = Pt(
    Fq(0x79be667ef9dcbbac55a06295ce870b07029bfcdb2dce28d959f2815b16f81798),
    Fq(0x483ada7726a3c4655da4fbfc0e1108a8fd17b448a68554199c47d08ffb10d4b8),
)

if __name__ == '__main__':
    p = G * Fr(42)
    q = G * Fr(24)
    r = Pt(p.x, -p.y)
    assert p + q == G * Fr(66)
    assert p + p == G * Fr(84)
    assert p - q == G * Fr(18)
    assert r == -p
    assert p + r == I
    assert p + I == p
    assert p * Fr(42) == G * Fr(1764)



import pyckb


def test_get_cells():
    pyckb.config.upgrade('http://127.0.0.1:8114')
    pyckb.config.current = pyckb.config.develop
    prikey = pyckb.core.PriKey(1)
    pubkey = prikey.pubkey()
    args = pyckb.core.hash(pubkey.sec())[:20].hex()
    script = pyckb.core.Script(
        pyckb.config.current.script.secp256k1_blake160.code_hash,
        pyckb.config.current.script.secp256k1_blake160.hash_type,
        bytearray.fromhex(args)
    )
    search = {'script': script.json(), 'script_type': 'lock'}
    assert pyckb.rpc.get_cells(search, 'asc', '0xff', None)['objects'] != []


def test_get_cells_capacity():
    pyckb.config.upgrade('http://127.0.0.1:8114')
    pyckb.config.current = pyckb.config.develop
    prikey = pyckb.core.PriKey(1)
    pubkey = prikey.pubkey()
    args = pyckb.core.hash(pubkey.sec())[:20].hex()
    script = pyckb.core.Script(
        pyckb.config.current.script.secp256k1_blake160.code_hash,
        pyckb.config.current.script.secp256k1_blake160.hash_type,
        bytearray.fromhex(args)
    )
    search = {'script': script.json(), 'script_type': 'lock'}
    assert int(pyckb.rpc.get_cells_capacity(search)['capacity'], 16) >= 0


def test_get_current_epoch():
    pyckb.config.upgrade('http://127.0.0.1:8114')
    pyckb.config.current = pyckb.config.develop
    assert int(pyckb.rpc.get_current_epoch()['number'], 0) >= 0


def test_get_header_by_number():
    pyckb.config.upgrade('http://127.0.0.1:8114')
    pyckb.config.current = pyckb.config.develop
    header_json = pyckb.rpc.get_header_by_number('0x1')
    header = pyckb.core.Header.json_decode(header_json)
    assert header.raw.number == 1
    assert header.hash() == bytearray.fromhex(header_json['hash'][2:])


def test_get_indexer_tip():
    pyckb.config.upgrade('http://127.0.0.1:8114')
    pyckb.config.current = pyckb.config.develop
    assert int(pyckb.rpc.get_indexer_tip()['block_number'], 16) >= 0


def test_get_tip_block_numner():
    pyckb.config.upgrade('http://127.0.0.1:8114')
    pyckb.config.current = pyckb.config.develop
    assert int(pyckb.rpc.get_tip_block_number(), 16) >= 0



import argparse
import pyckb

# Deploy a script to the chain.

parser = argparse.ArgumentParser()
parser.add_argument('--file', type=str, help='script file')
parser.add_argument('--net', type=str, choices=['develop', 'mainnet', 'testnet'], default='develop')
parser.add_argument('--prikey', type=str, help='private key')
args = parser.parse_args()

if args.net == 'develop':
    pyckb.config.upgrade('http://127.0.0.1:8114')
    pyckb.config.current = pyckb.config.develop
if args.net == 'mainnet':
    pyckb.config.current = pyckb.config.mainnet
if args.net == 'testnet':
    pyckb.config.current = pyckb.config.testnet

user = pyckb.wallet.Wallet(int(args.prikey, 0))
hole = pyckb.core.Script(
    pyckb.config.current.script.secp256k1_blake160.code_hash,
    pyckb.config.current.script.secp256k1_blake160.hash_type,
    bytearray([0] * 20)
)

with open(args.file, 'rb') as f:
    data = f.read()
    print(f'script.code_hash = 0x{pyckb.core.hash(data).hex()}')
    print(f'script.hash_type = 2(data1)')
    hash = user.script_deploy(hole, data)
    print(f'out_point.hash   = 0x{hash.hex()}')
    print(f'out_point.index  = 0')
    pyckb.rpc.wait(f'0x{hash.hex()}')



import argparse
import pyckb

# Get the capacity by an address.

parser = argparse.ArgumentParser()
parser.add_argument('--addr', type=str, help='address')
parser.add_argument('--net', type=str, choices=['develop', 'mainnet', 'testnet'], default='develop')
args = parser.parse_args()

if args.net == 'develop':
    pyckb.config.upgrade('http://127.0.0.1:8114')
    pyckb.config.current = pyckb.config.develop
if args.net == 'mainnet':
    pyckb.config.current = pyckb.config.mainnet
if args.net == 'testnet':
    pyckb.config.current = pyckb.config.testnet

capacity = int(pyckb.rpc.get_cells_capacity({
    'script': pyckb.core.address_decode(args.addr).json(),
    'script_type': 'lock',
    'filter': {
        'script_len_range': ['0x0', '0x1']
    }
})['capacity'], 16)
print(capacity / pyckb.denomination.ckbytes)



import argparse
import json
import pyckb

# Dump full transaction data for [ckb-debugger](https://github.com/nervosnetwork/ckb-standalone-debugger) to use.

parser = argparse.ArgumentParser()
parser.add_argument('--file', type=str, help='transaction file')
parser.add_argument('--hash', type=str, help='transaction hash')
parser.add_argument('--net', type=str, choices=['develop', 'mainnet', 'testnet'], default='develop')
args = parser.parse_args()

assert args.file or args.hash

if args.net == 'develop':
    pyckb.config.upgrade('http://127.0.0.1:8114')
    pyckb.config.current = pyckb.config.develop
if args.net == 'mainnet':
    pyckb.config.current = pyckb.config.mainnet
if args.net == 'testnet':
    pyckb.config.current = pyckb.config.testnet

if args.file:
    tx_json = json.load(open(args.file))
if args.hash:
    tx_json = pyckb.rpc.get_transaction(args.hash)['transaction']
tx = pyckb.core.Transaction.json_decode(tx_json)

tx_mock = {
    'mock_info': {
        'inputs': [],
        'cell_deps': [],
        'header_deps': []
    },
    'tx': tx.json(),
}

for i in tx.raw.inputs:
    i_tx_rpc = pyckb.rpc.get_transaction(f'0x{i.previous_output.tx_hash.hex()}')
    i_tx = pyckb.core.Transaction.json_decode(i_tx_rpc['transaction'])
    i_output = i_tx.raw.outputs[i.previous_output.index]
    i_data = i_tx.raw.outputs_data[i.previous_output.index]
    i_header = i_tx_rpc['tx_status']['block_hash']
    tx_mock['mock_info']['inputs'].append({
        'input': i.json(),
        'output': i_output.json(),
        'data': f'0x{i_data.hex()}',
        'header': i_header,
    })
for c in tx.raw.cell_deps:
    c_tx_rpc = pyckb.rpc.get_transaction(f'0x{c.out_point.tx_hash.hex()}')
    c_tx = pyckb.core.Transaction.json_decode(c_tx_rpc['transaction'])
    c_output = c_tx.raw.outputs[c.out_point.index]
    c_data = c_tx.raw.outputs_data[c.out_point.index]
    c_header = c_tx_rpc['tx_status']['block_hash']
    tx_mock['mock_info']['cell_deps'].append({
        'cell_dep': c.json(),
        'output': c_output.json(),
        'data': f'0x{c_data.hex()}',
        'header': c_header,
    })
for i in range(len(tx.raw.cell_deps)):
    c = tx.raw.cell_deps[i]
    if c.dep_type != 1:
        continue
    c_data = bytearray.fromhex(tx_mock['mock_info']['cell_deps'][i]['data'][2:])
    for e in pyckb.molecule.decode_fixvec(c_data):
        d = pyckb.core.OutPoint.molecule_decode(e)
        d_tx_rpc = pyckb.rpc.get_transaction(f'0x{d.tx_hash.hex()}')
        d_tx = pyckb.core.Transaction.json_decode(d_tx_rpc['transaction'])
        d_output = d_tx.raw.outputs[d.index]
        d_data = d_tx.raw.outputs_data[d.index]
        d_header = d_tx_rpc['tx_status']['block_hash']
        tx_mock['mock_info']['cell_deps'].append({
            'cell_dep': pyckb.core.CellDep(d, 0).json(),
            'output': d_output.json(),
            'data': f'0x{d_data.hex()}',
            'header': d_header,
        })
for h in tx.raw.header_deps:
    h_rpc = pyckb.rpc.get_header(f'0x{h.hex()}')
    tx_mock['mock_info']['header_deps'].push(h_rpc)

print(json.dumps(tx_mock, indent=4))



import argparse
import math
import pyckb

# Attempt to withdraw all funds from Dao. When running the test case of pyckb by 'pytest -v', a part of ckb will be
# locked in Dao. Use this script to recover this part of the funds.

parser = argparse.ArgumentParser()
parser.add_argument('--net', type=str, choices=['develop', 'mainnet', 'testnet'], default='develop')
parser.add_argument('--prikey', type=str, help='private key')
args = parser.parse_args()

if args.net == 'develop':
    pyckb.config.upgrade('http://127.0.0.1:8114')
    pyckb.config.current = pyckb.config.develop
if args.net == 'mainnet':
    pyckb.config.current = pyckb.config.mainnet
if args.net == 'testnet':
    pyckb.config.current = pyckb.config.testnet

user = pyckb.wallet.Wallet(int(args.prikey, 0))
for e in user.dao_livecell():
    if e['output_data'] == '0x0000000000000000':
        out_point = pyckb.core.OutPoint.json_decode(e['out_point'])
        hash = user.dao_prepare(out_point)
        pyckb.rpc.wait(f'0x{hash.hex()}')
        print(f'0x{hash.hex()}')
    else:
        deposit_block_number_byte = bytearray.fromhex(e['output_data'][2:])
        deposit_block_number = int.from_bytes(deposit_block_number_byte, 'little')
        deposit_block_header = pyckb.rpc.get_header_by_number(hex(deposit_block_number))
        deposit_block_epoch = pyckb.core.epoch_decode(int(deposit_block_header['epoch'], 16))
        deposit_block_epoch_float = deposit_block_epoch[0] + deposit_block_epoch[1] / deposit_block_epoch[2]
        prepare_block_header = pyckb.rpc.get_header_by_number(e['block_number'])
        prepare_block_epoch = pyckb.core.epoch_decode(int(prepare_block_header['epoch'], 16))
        prepare_block_epoch_float = prepare_block_epoch[0] + prepare_block_epoch[1] / prepare_block_epoch[2]
        extract_since_delay = math.ceil((prepare_block_epoch_float - deposit_block_epoch_float) / 180) * 180
        extract_since_epoch_float = deposit_block_epoch_float + extract_since_delay
        current_block_epoch = pyckb.core.epoch_decode(int(pyckb.rpc.get_tip_header()['epoch'], 16))
        current_block_epoch_float = current_block_epoch[0] + current_block_epoch[1] / current_block_epoch[2]
        if current_block_epoch_float < extract_since_epoch_float:
            break
        out_point = pyckb.core.OutPoint.json_decode(e['out_point'])
        hash = user.dao_extract(out_point)
        pyckb.rpc.wait(f'0x{hash.hex()}')
        print(f'0x{hash.hex()}')



import argparse
import pyckb
import itertools
import random
import requests
import time

# This script generates a random address to request ckb from the faucet, and transfers the obtained ckb to the
# specified user, breaking through the single address limit of the faucet.
#
# https://faucet.nervos.org/
pyckb.config.current = pyckb.config.testnet
parser = argparse.ArgumentParser()
parser.add_argument('--addr', type=str, help='ckb address')
args = parser.parse_args()

kana = pyckb.wallet.Wallet(random.randint(0, pyckb.secp256k1.N - 1))
resp = requests.post('https://faucet-api.nervos.org/claim_events', json={
    'claim_event': {
        'address_hash': kana.addr,
        'amount': '300000'
    }
})
assert resp.status_code == 200, resp.text
id = resp.json()['data']['id']

for _ in itertools.repeat(0):
    time.sleep(1)
    resp = requests.get(f'https://faucet-api.nervos.org/claim_events')
    assert resp.status_code == 200
    data = resp.json()['claimEvents']['data']
    data = [e for e in data if e['id'] == id]
    if not data:
        continue
    hash = data[0]['attributes']['txHash']
    if not hash:
        continue
    break

print(f'hash: {hash}')
pyckb.rpc.wait(hash)
hash = kana.transfer_all(pyckb.core.address_decode(args.addr))
print(f'hash: 0x{hash.hex()}')
pyckb.rpc.wait(f'0x{hash.hex()}')



# Copyright (c) 2017, 2020 Pieter Wuille
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# BIP-0173 https://github.com/bitcoin/bips/blob/master/bip-0173.mediawiki
# BIP-0350 https://github.com/bitcoin/bips/blob/master/bip-0350.mediawiki
# Derived from https://raw.githubusercontent.com/sipa/bech32/master/ref/python/segwit_addr.py
#
# Reference implementation for Bech32/bech32 and segwit addresses.

import typing

CHARSET = 'qpzry9x8gf2tvdw0s3jn54khce6mua7l'
CONST_0 = 1
CONST_M = 0x2bc830a3


def bech32_polymod(data: bytearray) -> int:
    # Internal function that computes the Bech32 checksum.
    assert isinstance(data, bytearray)
    gen = [0x3b6a57b2, 0x26508e6d, 0x1ea119fa, 0x3d4233dd, 0x2a1462b3]
    chk = 1
    for val in data:
        top = chk >> 25
        chk = chk & 0x1ffffff
        chk = chk << 5 ^ val
        for i in range(5):
            chk ^= gen[i] if ((top >> i) & 1) else 0
    return chk


def bech32_hrpconv(hrp: str) -> bytearray:
    # Expand the HRP into values for checksum computation.
    r = bytearray()
    r.extend([ord(x) >> 5 for x in hrp])
    r.append(0)
    r.extend([ord(x) & 31 for x in hrp])
    return r


def bech32_re_arrange_5(data: bytearray) -> bytearray:
    # Re-arrange those bits into groups of 5, and pad with zeroes at the end if needed.
    assert isinstance(data, bytearray)
    acc = 0
    bit = 0
    ret = bytearray()
    max_val = 0x1f
    max_acc = 0xfff
    for b in data:
        acc = ((acc << 8) | b) & max_acc
        bit += 8
        for _ in range(bit // 5):
            bit -= 5
            ret.append((acc >> bit) & max_val)
    if bit:
        pad_bit = 5 - bit
        ret.append((acc << pad_bit) & max_val)
    return ret


def bech32_create_checksum(hrp: str, ver: int, data: bytearray) -> bytearray:
    hrpdata = bech32_hrpconv(hrp) + data
    polymod = bech32_polymod(hrpdata + bytearray(6))
    if ver == 0:
        polymod = polymod ^ CONST_0
    if ver >= 1:
        polymod = polymod ^ CONST_M
    return bytearray([(polymod >> 5 * (5 - i)) & 31 for i in range(6)])


def bech32_verify_checksum(hrp: str, ver: int, data: bytearray) -> bool:
    if ver == 0:
        return bech32_polymod(bech32_hrpconv(hrp) + bytearray(data)) == CONST_0
    if ver >= 1:
        return bech32_polymod(bech32_hrpconv(hrp) + bytearray(data)) == CONST_M


def bech32_decode(ver: int, bech: str) -> typing.Tuple[str, bytearray]:
    # Validate a string, and determine HRP and data.
    for c in bech:
        assert ord(c) >= ord('!')
        assert ord(c) <= ord('~')
    bech = bech.lower()
    pos = bech.rfind('1')
    assert pos > 0
    assert pos + 6 < len(bech)
    for c in bech[pos+1:]:
        assert c in CHARSET
    hrp = bech[:pos]
    data = bytearray([CHARSET.find(x) for x in bech[pos+1:]])
    assert bech32_verify_checksum(hrp, ver, data)
    return hrp, data[:-6]


def bech32_encode(hrp: str, ver: int, data: bytearray) -> str:
    datasum = data + bech32_create_checksum(hrp, ver, data)
    return hrp + '1' + ''.join([CHARSET[d] for d in datasum])


def bech32_re_arrange_8(data: bytearray) -> bytearray:
    # Re-arrange those bits into groups of 8 bits. Any incomplete group at the end MUST be 4 bits or less, MUST be all
    # zeroes, and is discarded.
    assert isinstance(data, bytearray)
    acc = 0
    bit = 0
    ret = bytearray()
    max_val = 0xff
    max_acc = 0xfff
    for b in data:
        assert b <= 0x1f
        acc = ((acc << 5) | b) & max_acc
        bit += 5
        for _ in range(bit // 8):
            bit -= 8
            ret.append((acc >> bit) & max_val)
    assert bit < 5
    return ret


def decode(hrp: str, addr: str) -> bytearray:
    # Decode a segwit address.
    pre, data = bech32_decode(1, addr)
    assert pre == hrp
    return bech32_re_arrange_8(data)


def encode(hrp: str, prog: bytearray) -> str:
    # Encode a segwit address.
    assert isinstance(prog, bytearray)
    r = bech32_encode(hrp, 1, bech32_re_arrange_5(prog))
    assert prog == decode(hrp, r)
    return r
